import os
import glob
import shutil
import argparse
import pandas as pd
from Bio import SeqIO
from time import sleep
import multiprocessing as mp
from datetime import datetime
from distutils.spawn import find_executable


def sep_path_basename_ext(file_in):

    # separate path and file name
    file_path, file_name = os.path.split(file_in)
    if file_path == '':
        file_path = '.'

    # separate file basename and extension
    file_basename, file_extension = os.path.splitext(file_name)

    return file_path, file_basename, file_extension


def get_max_value_in_dict(dict_in):
    max_value = 0
    max_value_key = ''
    for each_key in dict_in:
        if dict_in[each_key] > max_value:
            max_value = dict_in[each_key]
            max_value_key = each_key
    return max_value, max_value_key


def get_total_value_in_dict(dict_in):
    total_value = 0
    for each_key in dict_in:
        total_value += dict_in[each_key]
    return total_value


def cigar_splitter(cigar):

    # get the position of letters
    letter_pos_list = []
    n = 0
    for each_element in cigar:
        if each_element.isalpha() is True:
            letter_pos_list.append(n)
        n += 1

    # split cigar
    index = 0
    cigar_splitted = []
    while index <= len(letter_pos_list) - 1:
        if index == 0:
            cigar_splitted.append(cigar[:(letter_pos_list[index] + 1)])
        else:
            cigar_splitted.append(cigar[(letter_pos_list[index - 1] + 1):(letter_pos_list[index] + 1)])
        index += 1

    return cigar_splitted


def sort_csv_by_col(file_in, file_out, col_header):
    df_in        = pd.read_csv(file_in)
    df_in_sorted = df_in.sort_values(by=[col_header], ascending=False)
    df_in_sorted.to_csv(file_out, index=False)


def blast_results_to_pairwise_16s_iden_dict(blastn_output, align_len_cutoff, cov_cutoff):

    pairwise_iden_dict = {}
    for match in open(blastn_output):
        match_split = match.strip().split('\t')
        query = match_split[0]
        subject = match_split[1]
        iden = float(match_split[2])
        align_len = int(match_split[3])
        query_len = int(match_split[12])
        subject_len = int(match_split[13])
        coverage_q = float(align_len) * 100 / float(query_len)
        coverage_s = float(align_len) * 100 / float(subject_len)

        if (align_len >= align_len_cutoff) and (query != subject) and (coverage_q >= cov_cutoff) and (coverage_s >= cov_cutoff):
            query_to_subject_key = '__|__'.join(sorted([query, subject]))
            if query_to_subject_key not in pairwise_iden_dict:
                pairwise_iden_dict[query_to_subject_key] = iden
            else:
                if iden > pairwise_iden_dict[query_to_subject_key]:
                    pairwise_iden_dict[query_to_subject_key] = iden

    return pairwise_iden_dict


def filter_linkages_iteratively(file_in, sort_by_col_header, pairwise_16s_iden_dict, genomic_seq_depth_dict, marker_gene_depth_dict, min_16s_gnm_multiple, within_genome_16s_divergence_cutoff, min_linkages, file_out):

    file_in_path, file_in_basename, file_in_extension = sep_path_basename_ext(file_in)
    file_in_sorted = '%s/%s_sorted%s' % (file_in_path, file_in_basename, file_in_extension)

    # sort file in
    sort_csv_by_col(file_in, file_in_sorted, sort_by_col_header)

    # fileter linkage
    file_out_handle = open(file_out, 'w')
    MarkerGene_with_assignment = set()
    GenomicSeq_best_marker_dict = {}
    for each_match in open(file_in_sorted):
        if each_match.startswith('MarkerGene,GenomicSeq,Number'):
            file_out_handle.write(each_match)
        else:
            match_split = each_match.strip().split(',')
            MarkerGene = match_split[0][12:]
            GenomicSeq = match_split[1][12:]
            linkage_num = int(match_split[2])

            if linkage_num >= min_linkages:

                # consider depth
                if min_16s_gnm_multiple > 0:
                    MarkerGene_depth = marker_gene_depth_dict[MarkerGene]
                    GenomicSeq_depth = genomic_seq_depth_dict[GenomicSeq]
                    if (MarkerGene_depth/GenomicSeq_depth) >= min_16s_gnm_multiple:
                        if MarkerGene not in MarkerGene_with_assignment:

                            if GenomicSeq not in GenomicSeq_best_marker_dict:
                                GenomicSeq_best_marker_dict[GenomicSeq] = MarkerGene
                                file_out_handle.write(each_match)
                                MarkerGene_with_assignment.add(MarkerGene)
                            else:
                                current_GenomicSeq_best_marker = GenomicSeq_best_marker_dict[GenomicSeq]
                                key_str = '__|__'.join(sorted([MarkerGene, current_GenomicSeq_best_marker]))

                                iden_with_best_marker = 0
                                if key_str in pairwise_16s_iden_dict:
                                    iden_with_best_marker = pairwise_16s_iden_dict[key_str]

                                if iden_with_best_marker >= within_genome_16s_divergence_cutoff:
                                    file_out_handle.write(each_match)
                                    MarkerGene_with_assignment.add(MarkerGene)
                # ignore depth
                else:
                    if MarkerGene not in MarkerGene_with_assignment:
                        if GenomicSeq not in GenomicSeq_best_marker_dict:
                            GenomicSeq_best_marker_dict[GenomicSeq] = MarkerGene
                            file_out_handle.write(each_match)
                            MarkerGene_with_assignment.add(MarkerGene)
                        else:
                            current_GenomicSeq_best_marker = GenomicSeq_best_marker_dict[GenomicSeq]
                            key_str = '__|__'.join(sorted([MarkerGene, current_GenomicSeq_best_marker]))

                            iden_with_best_marker = 0
                            if key_str in pairwise_16s_iden_dict:
                                iden_with_best_marker = pairwise_16s_iden_dict[key_str]

                            if iden_with_best_marker >= within_genome_16s_divergence_cutoff:
                                file_out_handle.write(each_match)
                                MarkerGene_with_assignment.add(MarkerGene)

    file_out_handle.close()

    # remove tmp file
    # os.remove(file_in_sorted)


def get_best_ctg_or_16s_for_gap_seq_iteratively(file_in, sort_by_col_header, min_linkages, file_out):

    file_in_path, file_in_basename, file_in_extension = sep_path_basename_ext(file_in)
    file_in_sorted = '%s/%s_sorted%s' % (file_in_path, file_in_basename, file_in_extension)

    # sort file in
    sort_csv_by_col(file_in, file_in_sorted, sort_by_col_header)

    # fileter linkage
    file_out_handle = open(file_out, 'w')
    gap_seq_with_assignment = set()
    for each_match in open(file_in_sorted):
        if each_match.startswith('Gap_seq,'):
            file_out_handle.write(each_match)
        else:
            match_split = each_match.strip().split(',')
            gap_seq_id = match_split[0]
            linkage_num = int(match_split[2])
            if (linkage_num >= min_linkages) and (gap_seq_id not in gap_seq_with_assignment):
                file_out_handle.write(each_match)
                gap_seq_with_assignment.add(gap_seq_id)
    file_out_handle.close()

    # remove tmp file
    # os.remove(file_in_sorted)


################################################### files/parameters ###################################################

wd = '/Users/songweizhi/Desktop/GapFilling_all'
wd = '/Users/songweizhi/Desktop/GapFilling_sep'

free_living_mate_gnm = '%s/free_living_mate_ctg.txt'                                        % wd
free_living_mate_16s = '%s/free_living_mate_16s.txt'                                        % wd
sam_file             = '%s/scaffolds.sam'                                                   % wd

gnm_ctg_connector    = '___'
dict_key_connector   = '__|__'
min_read_num         = 3

if wd.endswith('_all'):
    blastdb_16s = '%s/MBARC26_all_depth_assemblies.dereplicated_99.5_500bp.fasta'           % wd
else:
    blastdb_16s = '%s/rest_16S.fasta'                                                       % wd


########################################################################################################################

# define output file name
blast_results_all_vs_all_16s   = '%s/16s_all_vs_all.txt'                                    % wd

stats_GapFilling_file_16s          = '%s/stats_GapFilling_16s.txt'                          % wd
stats_GapFilling_file_ctg          = '%s/stats_GapFilling_ctg.txt'                          % wd
stats_GapFilling_file_gnm          = '%s/stats_GapFilling_gnm.txt'                          % wd

stats_GapFilling_file_filtered_16s = '%s/stats_GapFilling_16s_filtered.txt'                 % wd
stats_GapFilling_file_filtered_ctg = '%s/stats_GapFilling_ctg_filtered.txt'                 % wd
stats_GapFilling_file_filtered_gnm = '%s/stats_GapFilling_gnm_filtered.txt'                 % wd

stats_GapFilling_file              = '%s/stats_GapFilling.txt'                              % wd
stats_GapFilling_file_filtered     = '%s/stats_GapFilling_filtered.txt'                     % wd

########################################################################################################################

reads_to_extract_to_ref_dict_gnm = {}
for i in open(free_living_mate_gnm):
    i_split = i.strip().split('\t')
    reads_to_extract_to_ref_dict_gnm[i_split[0]] = i_split[1]

reads_to_extract_to_ref_dict_16s = {}
for j in open(free_living_mate_16s):
    j_split = j.strip().split('\t')
    reads_to_extract_to_ref_dict_16s[j_split[0]] = j_split[1]

gap_seq_to_reads_dict = {}
for each_line in open(sam_file):
    if not each_line.startswith('@'):
        each_line_split = each_line.strip().split('\t')
        read_id         = each_line_split[0]
        read_id_base    = '.'.join(read_id.split('.')[:-1])
        read_strand     = read_id.split('.')[-1]
        ref_id          = each_line_split[2]
        ref_pos         = int(each_line_split[3])
        cigar           = each_line_split[5]
        cigar_splitted  = cigar_splitter(cigar)
        if (len(cigar_splitted) == 1) and (cigar[-1] == 'M'):
            if ref_id not in gap_seq_to_reads_dict:
                gap_seq_to_reads_dict[ref_id] = [read_id]
            else:
                gap_seq_to_reads_dict[ref_id].append(read_id)


stats_GapFilling_file_16s_handle = open(stats_GapFilling_file_16s, 'w')
stats_GapFilling_file_ctg_handle = open(stats_GapFilling_file_ctg, 'w')
stats_GapFilling_file_16s_handle.write('Gap_seq,s16,Number\n')
stats_GapFilling_file_ctg_handle.write('Gap_seq,ctg,Number\n')
for gap_seq in gap_seq_to_reads_dict:
    gap_seq_mapped_reads = gap_seq_to_reads_dict[gap_seq]
    gap_seq_mapped_reads_linkages_16s = {}
    gap_seq_mapped_reads_linkages_ctg = {}
    gap_seq_mapped_reads_linkages_gnm = {}
    for mapped_read in gap_seq_mapped_reads:

        # get mate ref
        mapped_read_mate_ref = ''
        mapped_read_mate_ref_gnm = ''
        mate_ref_is_16s = False
        mate_ref_is_ctg = False
        if mapped_read in reads_to_extract_to_ref_dict_gnm:
            mapped_read_mate_ref     = reads_to_extract_to_ref_dict_gnm[mapped_read]
            mapped_read_mate_ref_gnm = mapped_read_mate_ref.split(gnm_ctg_connector)[0]
            mate_ref_is_ctg = True
        if mapped_read in reads_to_extract_to_ref_dict_16s:
            mapped_read_mate_ref = reads_to_extract_to_ref_dict_16s[mapped_read]
            mate_ref_is_16s = True

        # store 16s mate in dict
        if (mate_ref_is_16s is True) and (mate_ref_is_ctg is False):
            if mapped_read_mate_ref not in gap_seq_mapped_reads_linkages_16s:
                gap_seq_mapped_reads_linkages_16s[mapped_read_mate_ref] = 1
            else:
                gap_seq_mapped_reads_linkages_16s[mapped_read_mate_ref] += 1

        # store ctg mate in dict
        if (mate_ref_is_16s is False) and (mate_ref_is_ctg is True):

            if mapped_read_mate_ref not in gap_seq_mapped_reads_linkages_ctg:
                gap_seq_mapped_reads_linkages_ctg[mapped_read_mate_ref] = 1
            else:
                gap_seq_mapped_reads_linkages_ctg[mapped_read_mate_ref] += 1

            if mapped_read_mate_ref_gnm not in gap_seq_mapped_reads_linkages_gnm:
                gap_seq_mapped_reads_linkages_gnm[mapped_read_mate_ref_gnm] = 1
            else:
                gap_seq_mapped_reads_linkages_gnm[mapped_read_mate_ref_gnm] += 1

    for each_16s in gap_seq_mapped_reads_linkages_16s:
        stats_GapFilling_file_16s_handle.write('%s,%s,%s\n' % (gap_seq, each_16s, gap_seq_mapped_reads_linkages_16s[each_16s]))

    for each_ctg in gap_seq_mapped_reads_linkages_ctg:
        stats_GapFilling_file_ctg_handle.write('%s,%s,%s\n' % (gap_seq, each_ctg, gap_seq_mapped_reads_linkages_ctg[each_ctg]))

stats_GapFilling_file_16s_handle.close()
stats_GapFilling_file_ctg_handle.close()


get_best_ctg_or_16s_for_gap_seq_iteratively(stats_GapFilling_file_16s, 'Number', min_read_num, stats_GapFilling_file_filtered_16s)
get_best_ctg_or_16s_for_gap_seq_iteratively(stats_GapFilling_file_ctg, 'Number', min_read_num, stats_GapFilling_file_filtered_ctg)


gap_seq_to_16s_dict = {}
for each_line in open(stats_GapFilling_file_filtered_16s):
    if not each_line.startswith('Gap_seq,'):
        each_line_split = each_line.strip().split(',')
        gap_seq_id = each_line_split[0]
        s16_id = each_line_split[1]
        link_num = int(each_line_split[2])
        gap_seq_to_16s_dict[gap_seq_id] = {s16_id: link_num}


gap_seq_to_gnm_dict = {}
for each_match in open(stats_GapFilling_file_filtered_ctg):
    if not each_match.startswith('Gap_seq,'):
        match_split = each_match.strip().split(',')
        gap_seq_id = match_split[0]
        ctg_id = match_split[1]
        ctg_gnm = ctg_id.split(gnm_ctg_connector)[0]
        linkage_num = int(match_split[2])
        gap_seq_to_gnm_dict[gap_seq_id] = {ctg_gnm: linkage_num}


gnm_to_16s_linkage_dict = {}
for gap_seq in gap_seq_to_gnm_dict:
    if gap_seq in gap_seq_to_16s_dict:
        current_gap_seq_matched_gnm          = list(gap_seq_to_gnm_dict[gap_seq].items())[0][0]
        current_gap_seq_matched_gnm_link_num = list(gap_seq_to_gnm_dict[gap_seq].items())[0][1]
        current_gap_seq_matched_16s          = list(gap_seq_to_16s_dict[gap_seq].items())[0][0]
        current_gap_seq_matched_16s_link_num = list(gap_seq_to_16s_dict[gap_seq].items())[0][1]
        gnm_to_16s_key = '%s%s%s' % (current_gap_seq_matched_gnm, dict_key_connector, current_gap_seq_matched_16s)
        gnm_to_16s_link_num = current_gap_seq_matched_gnm_link_num + current_gap_seq_matched_16s_link_num
        if gnm_to_16s_key not in gnm_to_16s_linkage_dict:
            gnm_to_16s_linkage_dict[gnm_to_16s_key] = gnm_to_16s_link_num
        else:
            gnm_to_16s_linkage_dict[gnm_to_16s_key] += gnm_to_16s_link_num


stats_GapFilling_file_handle = open(stats_GapFilling_file, 'w')
stats_GapFilling_file_handle.write('MarkerGene,GenomicSeq,Number\n')
for gnm_to_16s in gnm_to_16s_linkage_dict:
    id_gnm = gnm_to_16s.split(dict_key_connector)[0]
    id_16s = gnm_to_16s.split(dict_key_connector)[1]
    stats_GapFilling_file_handle.write('MarkerGene__%s,GenomicSeq__%s,%s\n' % (id_16s, id_gnm, gnm_to_16s_linkage_dict[gnm_to_16s]))
stats_GapFilling_file_handle.close()


############################################# get pairwise_16s_iden_dict ##############################################

pwd_makeblastdb_exe = 'makeblastdb'
pwd_blastn_exe = 'blastn'
num_threads = 4
blast_parameters = '-evalue 1e-5 -outfmt "6 qseqid sseqid pident length mismatch gapopen qstart qend sstart send evalue bitscore qlen slen" -task blastn -num_threads %s' % num_threads

# makeblastdn with marker gene sequences
makeblastdb_16s_cmd = '%s -in %s -dbtype nucl -parse_seqids -logfile /dev/null' % (pwd_makeblastdb_exe, blastdb_16s)
os.system(makeblastdb_16s_cmd)

all_vs_all_16s_blastn_cmd = '%s -query %s -db %s -out %s %s' % (pwd_blastn_exe, blastdb_16s, blastdb_16s, blast_results_all_vs_all_16s, blast_parameters)
os.system(all_vs_all_16s_blastn_cmd)

pairwise_16s_iden_dict = blast_results_to_pairwise_16s_iden_dict(blast_results_all_vs_all_16s, 500, 90)


########################################################################################################################

filter_linkages_iteratively(stats_GapFilling_file, 'Number', pairwise_16s_iden_dict, {}, {}, 0, 98, 3, stats_GapFilling_file_filtered)





def cigar_splitter(cigar):

    # get the position of letters
    letter_pos_list = []
    n = 0
    for each_element in cigar:
        if (each_element.isalpha() is True) or (each_element == '='):
            letter_pos_list.append(n)
        n += 1

    # split cigar
    index = 0
    cigar_splitted = []
    while index <= len(letter_pos_list) - 1:
        if index == 0:
            cigar_splitted.append(cigar[:(letter_pos_list[index] + 1)])
        else:
            cigar_splitted.append(cigar[(letter_pos_list[index - 1] + 1):(letter_pos_list[index] + 1)])
        index += 1

    return cigar_splitted


def get_cigar_matched_and_mismatch_pct(cigar):

    cigar_splitted = cigar_splitter(cigar)

    total_len_with_s = 0
    total_len_without_s = 0
    matched_seq_len = 0
    mismatched_seq_len = 0
    for each_part in cigar_splitted:
        each_part_len = int(each_part[:-1])
        each_part_cate = each_part[-1]

        # get total len with/without s
        if each_part_cate in {'M', 'I', '=', 'X'}:
            total_len_with_s += each_part_len
            total_len_without_s += each_part_len

        # get total len with s
        if each_part_cate == 'S':
            total_len_with_s += each_part_len

        # get matched part len
        if each_part_cate == '=':
            matched_seq_len += each_part_len

        # get mismatched part len
        if each_part_cate in {'I', 'X'}:
            mismatched_seq_len += each_part_len

    matched_pct    = float("{0:.2f}".format(matched_seq_len*100/total_len_with_s))
    mismatch_pct = float("{0:.2f}".format(mismatched_seq_len*100/total_len_without_s))

    return matched_pct, mismatch_pct


def paired_blast_results_to_dict_by_mapping(unmapped_paired_reads_mapping_results):

    cigar_M_pct_min_value = 50
    mismatch_pct_max_value = 1

    unmapped_paired_reads_to_ctg_dict_by_mapping = {}

    ref_len_dict = {}
    for unmapped_read in open(unmapped_paired_reads_mapping_results):
        unmapped_read_split = unmapped_read.strip().split('\t')

        # get ref len dict
        if unmapped_read.startswith('@'):
            ref_id = ''
            ref_len = 0
            for each_element in unmapped_read_split:
                if each_element.startswith('SN:'):
                    ref_id = each_element[3:]
                if each_element.startswith('LN:'):
                    ref_len = int(each_element[3:])
            ref_len_dict[ref_id] = ref_len

        else:
            cigar = unmapped_read_split[5]
            if cigar != '*':
                unmapped_read_split = unmapped_read.strip().split('\t')
                read_id = unmapped_read_split[0]
                ref_id = unmapped_read_split[2]
                ref_len = ref_len_dict[ref_id]
                ref_id_with_prefix = 'GenomicSeq__%s' % unmapped_read_split[2]
                ref_pos = int(unmapped_read_split[3])
                read_seq = unmapped_read_split[9]
                read_len = len(read_seq)
                cigar_splitted = cigar_splitter(cigar)
                qualified_unmapped_read = False

                # e.g. 189=
                if ('=' in cigar) and (len(cigar_splitted) == 1):
                    qualified_unmapped_read = True

                elif len(cigar_splitted) == 2:

                    # e.g. ['44S', '176=']
                    if (cigar_splitted[0][-1] == 'S') and (cigar_splitted[1][-1] == '='):
                        cigar_M_pct = int(cigar_splitted[1][:-1]) * 100 / read_len
                        if (ref_pos == 1) and (cigar_M_pct >= cigar_M_pct_min_value):
                            qualified_unmapped_read = True

                    # e.g. ['154=', '66S']
                    if (cigar_splitted[0][-1] == '=') and (cigar_splitted[1][-1] == 'S'):
                        cigar_M_pct = int(cigar_splitted[0][:-1]) * 100 / read_len
                        matched_to_bp = ref_pos + int(cigar_splitted[0][:-1]) - 1
                        if (matched_to_bp == ref_len) and (cigar_M_pct >= cigar_M_pct_min_value):
                            qualified_unmapped_read = True


                elif len(cigar_splitted) == 3:

                    # e.g. ['134=', '1X', '22='], ['215=', '1X', '4=']
                    if (cigar_splitted[0][-1] == '=') and (cigar_splitted[2][-1] == '='):
                        mismatch_pct = int(cigar_splitted[1][:-1])*100/read_len
                        if mismatch_pct <= mismatch_pct_max_value:
                            qualified_unmapped_read = True

                elif len(cigar_splitted) == 4:

                    # e.g. ['118=', '1X', '99=', '1S'], ['8=', '1D', '127=', '84S']
                    if (cigar_splitted[0][-1] == '=') and (cigar_splitted[2][-1] == '=') and (cigar_splitted[3][-1] == 'S'):
                        cigar_M_pct = (int(cigar_splitted[0][:-1]) + int(cigar_splitted[2][:-1]))*100/read_len
                        if ((ref_pos + int(cigar_splitted[0][:-1]) + int(cigar_splitted[2][:-1])) == ref_len) and (int(cigar_splitted[1][:-1]) <= 2):
                            if cigar_M_pct >= cigar_M_pct_min_value:
                                qualified_unmapped_read = True

                    # e.g. ['51S', '158=', '1X', '10='] and ref_pos = 1
                    elif (cigar_splitted[0][-1] == 'S') and (cigar_splitted[1][-1] == '=') and (cigar_splitted[3][-1] == '='):
                        cigar_M_pct = (int(cigar_splitted[1][:-1]) + int(cigar_splitted[3][:-1]))*100/read_len
                        if (ref_pos == 1) and (int(cigar_splitted[2][:-1]) <= 1) and (cigar_M_pct >= cigar_M_pct_min_value):
                            qualified_unmapped_read = True

                # add to dict
                if qualified_unmapped_read is True:
                    if read_id not in unmapped_paired_reads_to_ctg_dict_by_mapping:
                        unmapped_paired_reads_to_ctg_dict_by_mapping[read_id] = [ref_id_with_prefix]
                    else:
                        unmapped_paired_reads_to_ctg_dict_by_mapping[read_id].append(ref_id_with_prefix)

    return unmapped_paired_reads_to_ctg_dict_by_mapping


def paired_blast_results_to_dict_by_mapping_new(unmapped_paired_reads_mapping_results):

    cigar_M_pct_min_value = 50
    mismatch_pct_max_value = 1

    unmapped_paired_reads_to_ctg_dict_by_mapping = {}

    for unmapped_read in open(unmapped_paired_reads_mapping_results):

        # get ref len dict
        if not unmapped_read.startswith('@'):
            unmapped_read_split = unmapped_read.strip().split('\t')
            cigar = unmapped_read_split[5]
            if cigar != '*':
                read_id = unmapped_read_split[0]
                ref_id_with_prefix = 'GenomicSeq__%s' % unmapped_read_split[2]
                cigar_match_pct, cigar_mismatch_pct = get_cigar_matched_and_mismatch_pct(cigar)

                if (cigar_match_pct >= cigar_M_pct_min_value) and (cigar_mismatch_pct <= mismatch_pct_max_value):

                    if read_id not in unmapped_paired_reads_to_ctg_dict_by_mapping:
                        unmapped_paired_reads_to_ctg_dict_by_mapping[read_id] = [ref_id_with_prefix]
                    else:
                        unmapped_paired_reads_to_ctg_dict_by_mapping[read_id].append(ref_id_with_prefix)

    return unmapped_paired_reads_to_ctg_dict_by_mapping



pwd_samfile_to_mag = '/Users/songweizhi/Desktop/BH_ER_050417_refined_bins_combined.sam'
unmapped_paired_reads_to_ctg_dict = paired_blast_results_to_dict_by_mapping(pwd_samfile_to_mag)
print(len(unmapped_paired_reads_to_ctg_dict))

unmapped_paired_reads_to_ctg_dict = paired_blast_results_to_dict_by_mapping_new(pwd_samfile_to_mag)
print(len(unmapped_paired_reads_to_ctg_dict))







def subset_prokaryotes_csv(file_in, id_file_in, prokaryotes_csv, file_out):

    num_to_id_dict = {}
    GTDB_genome_id_list = set()

    if id_file_in is True:
        for each_id in open(file_in):

            NCBI_ID_without_version = each_id.strip()
            if '.' in each_id.strip():
                NCBI_ID_without_version = '.'.join(each_id.strip().split('.')[:-1])

            NCBI_ID_no_version_only_number = NCBI_ID_without_version[4:]

            GTDB_genome_id_list.add(NCBI_ID_no_version_only_number)
            num_to_id_dict[NCBI_ID_no_version_only_number] = each_id.strip()

    else:
        for each in open(file_in):
            if not each.startswith('"ID",'):
                each_split = each.strip().split(',')

                NCBI_ID_with_version = each_split[0][1:-1]
                NCBI_ID_without_version = '.'.join(NCBI_ID_with_version.split('.')[:-1])
                NCBI_ID_no_version_only_number = NCBI_ID_without_version[4:]

                GTDB_genome_id_list.add(NCBI_ID_no_version_only_number)
                num_to_id_dict[NCBI_ID_no_version_only_number] = NCBI_ID_with_version


    GenBank_genomes_subset_handle = open(file_out, 'w')
    found_in_GenBank_genome_id_list = set()
    for GenBank_genome in open(prokaryotes_csv):
        if not GenBank_genome.startswith('#Organism Name'):
            GenBank_genome_split = GenBank_genome.strip().split(',')
            GenBank_genome_id_with_version = GenBank_genome_split[5][1:-1]
            GenBank_genome_id_without_version = '.'.join(GenBank_genome_id_with_version.split('.')[:-1])
            GenBank_genome_id_no_version_only_number = GenBank_genome_id_without_version[4:]
            if GenBank_genome_id_no_version_only_number in GTDB_genome_id_list:
                found_in_GenBank_genome_id_list.add(GenBank_genome_id_no_version_only_number)
                GenBank_genomes_subset_handle.write(GenBank_genome)
    GenBank_genomes_subset_handle.close()

    for each in GTDB_genome_id_list:
        if each not in found_in_GenBank_genome_id_list:
            print('not found: %s' % num_to_id_dict[each])


# file_in         = '/Users/songweizhi/Desktop/select_strain/GTDB-Pseudomonas_E.csv'
file_in         = '/Users/songweizhi/Desktop/accession_id.txt'
id_file_in      = True
prokaryotes_csv = '/Users/songweizhi/Desktop/select_strain/prokaryotes.csv'
file_out        = '/Users/songweizhi/Desktop/GenBank_g_test.csv'

subset_prokaryotes_csv(file_in, id_file_in, prokaryotes_csv, file_out)




import os
from Bio import SeqIO


def sep_path_basename_ext(file_in):

    # separate path and file name
    file_path, file_name = os.path.split(file_in)
    if file_path == '':
        file_path = '.'

    # separate file basename and extension
    file_basename, file_extension = os.path.splitext(file_name)

    return file_path, file_basename, file_extension


def cigar_splitter(cigar):

    # get the position of letters
    letter_pos_list = []
    n = 0
    for each_element in cigar:
        if (each_element.isalpha() is True) or (each_element == '='):
            letter_pos_list.append(n)
        n += 1

    # split cigar
    index = 0
    cigar_splitted = []
    while index <= len(letter_pos_list) - 1:
        if index == 0:
            cigar_splitted.append(cigar[:(letter_pos_list[index] + 1)])
        else:
            cigar_splitted.append(cigar[(letter_pos_list[index - 1] + 1):(letter_pos_list[index] + 1)])
        index += 1

    return cigar_splitted


def get_cigar_matched_and_mismatch_pct(cigar):

    cigar_splitted = cigar_splitter(cigar)

    total_len_with_s = 0
    total_len_without_s = 0
    matched_seq_len = 0
    mismatched_seq_len = 0
    for each_part in cigar_splitted:
        each_part_len = int(each_part[:-1])
        each_part_cate = each_part[-1]

        # get total len with/without s
        if each_part_cate in {'M', 'I', '=', 'X'}:
            total_len_with_s += each_part_len
            total_len_without_s += each_part_len

        # get total len with s
        if each_part_cate == 'S':
            total_len_with_s += each_part_len

        # get matched part len
        if each_part_cate == '=':
            matched_seq_len += each_part_len

        # get mismatched part len
        if each_part_cate in {'I', 'X', 'D'}:
            mismatched_seq_len += each_part_len

    matched_pct  = float("{0:.2f}".format(matched_seq_len*100/total_len_with_s))
    mismatch_pct = float("{0:.2f}".format(mismatched_seq_len*100/total_len_without_s))

    return matched_pct, mismatch_pct


# ref_in          = '/Users/songweizhi/Desktop/step_1_unlinked_combined_gnms.fasta'
sam_file        = '/Users/songweizhi/Desktop/step_1_unlinked_combined_gnms_ends_3000bp.sam'
max_gap_to_end  = 300
minCigarM       = 50
end_seq_len     = 3000
perfect_match_min_cigar_M_pct = 70
global_max_mismatch_pct = 1


# parse sam file
all_mapped_reads = set()
qualified_reads_dict = {}
qualified_reads_to_ref_dict = {}
ref_in_sam_len_dict = {}
for each_line in open(sam_file):
    each_line_split = each_line.strip().split('\t')

    if each_line.startswith('@'):
        ref_id = ''
        ref_len = 0
        for each_element in each_line_split:
            if each_element.startswith('SN:'):
                ref_id = each_element[3:]
            if each_element.startswith('LN:'):
                ref_len = int(each_element[3:])
        ref_in_sam_len_dict[ref_id] = ref_len
    else:
        cigar = each_line_split[5]
        if cigar != '*':
            read_id = each_line_split[0]
            read_id_base = '.'.join(read_id.split('.')[:-1])
            read_strand = read_id.split('.')[-1]
            ref_id = each_line_split[2]
            ref_len = ref_in_sam_len_dict[ref_id]
            ref_pos = int(each_line_split[3])
            cigar_splitted = cigar_splitter(cigar)
            cigar_match_pct, cigar_mismatch_pct = get_cigar_matched_and_mismatch_pct(cigar)
            all_mapped_reads.add(read_id)

            # get aligned length
            aligned_len = 0
            for each_section in cigar_splitted:
                each_section_len = int(each_section[:-1])
                each_section_cate = each_section[-1]
                if each_section_cate in {'D', '=', 'X'}:
                    aligned_len += each_section_len

            # filter mapped reads
            qualified_mapping = False
            if (cigar_match_pct >= perfect_match_min_cigar_M_pct) and (cigar_mismatch_pct <= global_max_mismatch_pct):

                # check left end for contig's left end sequence
                if ref_id[-2:] == '_l':
                    if ref_pos <= max_gap_to_end:
                        qualified_mapping = True

                # check right end for contig's right end sequence
                elif ref_id[-2:] == '_r':
                    if cigar_splitted[0][-1] != 'S':
                        ref_pos_end = ref_pos + aligned_len
                        if (ref_len - ref_pos_end) <= max_gap_to_end:
                            qualified_mapping = True

                # check both ends for contigs without subset
                else:
                    if (ref_pos <= max_gap_to_end) or ((ref_len - ref_pos - aligned_len) <= max_gap_to_end):
                        qualified_mapping = True

            # proceed with qualified mappings
            if qualified_mapping is True:
                qualified_reads_to_ref_dict[read_id] = ref_id

                if read_id_base not in qualified_reads_dict:
                    qualified_reads_dict[read_id_base] = [read_strand]
                else:
                    qualified_reads_dict[read_id_base].append(read_strand)



reads_to_extract_to_ref_dict = {}
for qualified_read in qualified_reads_dict:
    read_strand = qualified_reads_dict[qualified_read]
    if len(read_strand) == 1:

        mapped_mate = ''
        mate_to_extract = ''
        if read_strand == ['1']:
            mapped_mate = '%s.1' % (qualified_read)
            mate_to_extract = '%s.2' % (qualified_read)
        if read_strand == ['2']:
            mapped_mate = '%s.2' % (qualified_read)
            mate_to_extract = '%s.1' % (qualified_read)

        if mate_to_extract not in all_mapped_reads:
            reads_to_extract_to_ref_dict[mate_to_extract] = qualified_reads_to_ref_dict[mapped_mate]


print(len(reads_to_extract_to_ref_dict))





def cigar_splitter(cigar):

    # get the position of letters
    letter_pos_list = []
    n = 0
    for each_element in cigar:
        if each_element.isalpha() is True:
            letter_pos_list.append(n)
        n += 1

    # split cigar
    index = 0
    cigar_splitted = []
    while index <= len(letter_pos_list) - 1:
        if index == 0:
            cigar_splitted.append(cigar[:(letter_pos_list[index] + 1)])
        else:
            cigar_splitted.append(cigar[(letter_pos_list[index - 1] + 1):(letter_pos_list[index] + 1)])
        index += 1

    return cigar_splitted


def paired_blast_results_to_dict(blastn_results, iden_cutoff, query_cov_cutoff):

    query_to_subject_list_dict = {}
    for blast_hit in open(blastn_results):
        blast_hit_split = blast_hit.strip().split('\t')
        query = blast_hit_split[0]
        subject = blast_hit_split[1]
        subject_with_prefix = 'GenomicSeq__%s' % subject
        iden = float(blast_hit_split[2])
        align_len = int(blast_hit_split[3])
        query_len = int(blast_hit_split[12])
        subject_len = int(blast_hit_split[13])
        coverage_q = float(align_len) * 100 / float(query_len)

        # for perfect hits
        if (iden >= iden_cutoff) and (coverage_q == 100):
            if query not in query_to_subject_list_dict:
                query_to_subject_list_dict[query] = [subject_with_prefix]
            else:
                query_to_subject_list_dict[query].append(subject_with_prefix)

        # for nearly perfect hits
        elif (iden >= iden_cutoff) and (query_cov_cutoff <= coverage_q < 100):
            s_l = sorted([int(blast_hit_split[8]), int(blast_hit_split[9])])[0]
            s_r = sorted([int(blast_hit_split[8]), int(blast_hit_split[9])])[1]
            subject_min_gap = min([s_l, (subject_len - s_r)])
            if subject_min_gap <= 5:
                if query not in query_to_subject_list_dict:
                    query_to_subject_list_dict[query] = [subject_with_prefix]
                else:
                    query_to_subject_list_dict[query].append(subject_with_prefix)

    return query_to_subject_list_dict


########################################################################################################################

pwd_samfile                     = '/Users/songweizhi/Desktop/tuning_wd/BH_ER_050417_assembled_16S_uclust_0.995.sam'
unmapped_paired_reads_blastn    = '/Users/songweizhi/Desktop/tuning_wd/unmapped_paired_reads_blast.txt'
paired_reads_match_profile      = '/Users/songweizhi/Desktop/tuning_wd/match_profile_paired.txt'

reads_iden_cutoff               = 100
reads_cov_cutoff                = 90
perfect_match_min_cigar_M_pct   = 90
perfect_match_max_cigar_S_pct   = 10
genomic_seq_type                = 'mag'


pwd_samfile_to_mag                      = '/Users/songweizhi/Desktop/tuning_wd/BH_ER_050417_refined_bins_combined.sam'
paired_reads_match_profile_by_mapping   = '/Users/songweizhi/Desktop/tuning_wd/match_profile_paired_by_mapping.txt'


##################################################### extract reads ####################################################


# export clipping mapped reads and perfectly mapped reads
all_mapped_reads_set = set()
clipping_mapped_reads_list = set()
clipping_reads_mapped_part_dict = {}
perfectly_mapped_reads_dict = {}
for each_read in open(pwd_samfile):

    if not each_read.startswith('@'):
        each_read_split = each_read.strip().split('\t')
        read_id = each_read_split[0]
        read_id_base = '.'.join(read_id.split('.')[:-1])
        read_strand = read_id.split('.')[-1]
        ref_id = each_read_split[2]
        ref_id_with_prefix = 'MarkerGene__%s' % each_read_split[2]
        ref_pos = int(each_read_split[3])
        cigar = each_read_split[5]
        read_seq = each_read_split[9]
        cigar_splitted = cigar_splitter(cigar)
        read_id_with_ref_pos = '%s__x__%s__x__%s' % (read_id, ref_id, ref_pos)
        all_mapped_reads_set.add(read_id)

        # for perfectly mapped reads
        if ('M' in cigar) and (len(cigar_splitted) == 1):
            if read_id_base not in perfectly_mapped_reads_dict:
                perfectly_mapped_reads_dict[read_id_base] = {read_strand: [ref_id_with_prefix]}
            else:
                if read_strand not in perfectly_mapped_reads_dict[read_id_base]:
                    perfectly_mapped_reads_dict[read_id_base][read_strand] = [ref_id_with_prefix]
                else:
                    perfectly_mapped_reads_dict[read_id_base][read_strand].append(ref_id_with_prefix)

        if ('S' in cigar) and (len(cigar_splitted) == 2):
            cigar_M_len = 0
            cigar_S_len = 0
            if cigar_splitted[0][-1] == 'M':
                cigar_M_len = int(cigar_splitted[0][:-1])
                cigar_S_len = int(cigar_splitted[1][:-1])
            if cigar_splitted[1][-1] == 'M':
                cigar_M_len = int(cigar_splitted[1][:-1])
                cigar_S_len = int(cigar_splitted[0][:-1])

            cigar_M_pct = cigar_M_len * 100 / (cigar_M_len + cigar_S_len)
            cigar_S_pct = cigar_S_len * 100 / (cigar_M_len + cigar_S_len)

            # for clipping reads with unmapped part >= min_cigar_S
            if (cigar_M_pct >= 30) and (cigar_S_pct >= 30):
                read_seq_left = read_seq[: int(cigar_splitted[0][:-1])]
                read_seq_right = read_seq[-int(cigar_splitted[1][:-1]):]

                if cigar_splitted[0][-1] == 'M':

                    # write out the sequence of unmapped part

                    # store the match info of mapped part
                    if ('%s_l' % read_id_with_ref_pos) not in clipping_reads_mapped_part_dict:
                        clipping_reads_mapped_part_dict[('%s_l' % read_id_with_ref_pos)] = [ref_id_with_prefix]
                    else:
                        clipping_reads_mapped_part_dict[('%s_l' % read_id_with_ref_pos)].append(ref_id_with_prefix)

                if cigar_splitted[1][-1] == 'M':

                    # write out the sequence of unmapped part

                    # store the match info of mapped part
                    if ('%s_r' % read_id_with_ref_pos) not in clipping_reads_mapped_part_dict:
                        clipping_reads_mapped_part_dict[('%s_r' % read_id_with_ref_pos)] = [ref_id_with_prefix]
                    else:
                        clipping_reads_mapped_part_dict[('%s_r' % read_id_with_ref_pos)].append(ref_id_with_prefix)

                clipping_mapped_reads_list.add(read_id)

            # for clipping reads with unmapped part < min_cigar_S
            # treat these reads as perfectly mapped reads
            elif (cigar_M_pct >= perfect_match_min_cigar_M_pct) and (cigar_S_pct < perfect_match_max_cigar_S_pct):
                if read_id_base not in perfectly_mapped_reads_dict:
                    perfectly_mapped_reads_dict[read_id_base] = {read_strand: [ref_id_with_prefix]}
                else:
                    if read_strand not in perfectly_mapped_reads_dict[read_id_base]:
                        perfectly_mapped_reads_dict[read_id_base][read_strand] = [ref_id_with_prefix]
                    else:
                        perfectly_mapped_reads_dict[read_id_base][read_strand].append(ref_id_with_prefix)

        elif ('S' in cigar) and (len(cigar_splitted) > 2):

            treat_as_full_match = False
            if ((cigar_splitted[0][-1] == 'M') and (cigar_splitted[-1][-1] == 'S')) or (
                    (cigar_splitted[0][-1] == 'S') and (cigar_splitted[-1][-1] == 'M')):

                if len(cigar_splitted) == 4:

                    if (cigar_splitted[0][-1] == 'M') and (cigar_splitted[2][-1] == 'M') and (
                            cigar_splitted[3][-1] == 'S'):
                        mismatch_ratio = int(cigar_splitted[1][:-1]) * 100 / len(read_seq)
                        if mismatch_ratio <= 1:
                            cigar_M_pct = (int(cigar_splitted[0][:-1]) + int(cigar_splitted[2][:-1])) * 100 / len(
                                read_seq)
                            cigar_S_pct = (int(cigar_splitted[3][:-1])) * 100 / len(read_seq)
                            if (cigar_M_pct >= perfect_match_min_cigar_M_pct) and (
                                    cigar_S_pct < perfect_match_max_cigar_S_pct):
                                treat_as_full_match = True

                    if (cigar_splitted[0][-1] == 'S') and (cigar_splitted[1][-1] == 'M') and (
                            cigar_splitted[3][-1] == 'M'):
                        mismatch_ratio = int(cigar_splitted[2][:-1]) * 100 / len(read_seq)
                        if mismatch_ratio <= 1:
                            cigar_M_pct = (int(cigar_splitted[1][:-1]) + int(cigar_splitted[3][:-1])) * 100 / len(
                                read_seq)
                            cigar_S_pct = (int(cigar_splitted[0][:-1])) * 100 / len(read_seq)
                            if (cigar_M_pct >= perfect_match_min_cigar_M_pct) and (
                                    cigar_S_pct < perfect_match_max_cigar_S_pct):
                                treat_as_full_match = True

                elif len(cigar_splitted) == 6:

                    if (cigar_splitted[0][-1] == 'M') and (cigar_splitted[2][-1] == 'M') and (
                            cigar_splitted[4][-1] == 'M') and (cigar_splitted[5][-1] == 'S'):
                        mismatch_ratio = (int(cigar_splitted[1][:-1]) + int(cigar_splitted[3][:-1])) * 100 / len(
                            read_seq)
                        if mismatch_ratio <= 1:
                            cigar_M_pct = (int(cigar_splitted[0][:-1]) + int(cigar_splitted[2][:-1]) + int(
                                cigar_splitted[4][:-1])) * 100 / len(read_seq)
                            cigar_S_pct = (int(cigar_splitted[5][:-1])) * 100 / len(read_seq)
                            if (cigar_M_pct >= perfect_match_min_cigar_M_pct) and (
                                    cigar_S_pct < perfect_match_max_cigar_S_pct):
                                treat_as_full_match = True

                    if (cigar_splitted[0][-1] == 'S') and (cigar_splitted[1][-1] == 'M') and (
                            cigar_splitted[3][-1] == 'M') and (cigar_splitted[5][-1] == 'M'):
                        mismatch_ratio = (int(cigar_splitted[2][:-1]) + int(cigar_splitted[4][:-1])) * 100 / len(
                            read_seq)
                        if mismatch_ratio <= 1:
                            cigar_M_pct = (int(cigar_splitted[1][:-1]) + int(cigar_splitted[3][:-1]) + int(
                                cigar_splitted[5][:-1])) * 100 / len(read_seq)
                            cigar_S_pct = (int(cigar_splitted[0][:-1])) * 100 / len(read_seq)
                            if (cigar_M_pct >= perfect_match_min_cigar_M_pct) and (
                                    cigar_S_pct < perfect_match_max_cigar_S_pct):
                                treat_as_full_match = True

            if treat_as_full_match is True:
                if read_id_base not in perfectly_mapped_reads_dict:
                    perfectly_mapped_reads_dict[read_id_base] = {read_strand: [ref_id_with_prefix]}
                else:
                    if read_strand not in perfectly_mapped_reads_dict[read_id_base]:
                        perfectly_mapped_reads_dict[read_id_base][read_strand] = [ref_id_with_prefix]
                    else:
                        perfectly_mapped_reads_dict[read_id_base][read_strand].append(ref_id_with_prefix)

perfectly_mapped_read_singleton_dict = {}
for perfectly_mapped_read in perfectly_mapped_reads_dict:
    current_value = perfectly_mapped_reads_dict[perfectly_mapped_read]
    if len(current_value) == 1:
        perfectly_mapped_read_singleton_dict[perfectly_mapped_read] = current_value

######################################### parse blast results for paired reads #########################################

# filter blast results for paired reads
unmapped_paired_reads_to_ctg_dict = paired_blast_results_to_dict(unmapped_paired_reads_blastn, reads_iden_cutoff, reads_cov_cutoff)

def paired_blast_results_to_dict_by_mapping(unmapped_paired_reads_mapping_results):

    unmapped_paired_reads_to_ctg_dict_by_mapping = {}

    ref_len_dict = {}
    for unmapped_read in open(unmapped_paired_reads_mapping_results):

        # get ref len dict
        if unmapped_read.startswith('@'):
            unmapped_read_split = unmapped_read.strip().split('\t')
            ref_id = ''
            ref_len = 0
            for each_element in unmapped_read_split:
                if each_element.startswith('SN:'):
                    ref_id = each_element[3:]
                if each_element.startswith('LN:'):
                    ref_len = int(each_element[3:])
            ref_len_dict[ref_id] = ref_len

        else:
            qualified_unmapped_read = False
            unmapped_read_split = unmapped_read.strip().split('\t')
            read_id = unmapped_read_split[0]
            read_id_base = '.'.join(read_id.split('.')[:-1])
            read_strand = read_id.split('.')[-1]
            ref_id = unmapped_read_split[2]
            ref_len = ref_len_dict[ref_id]
            ref_id_with_prefix = 'GenomicSeq__%s' % unmapped_read_split[2]
            ref_pos = int(unmapped_read_split[3])
            cigar = unmapped_read_split[5]
            read_seq = unmapped_read_split[9]
            read_len = len(read_seq)
            cigar_splitted = cigar_splitter(cigar)

            # e.g. 189M
            if ('M' in cigar) and (len(cigar_splitted) == 1):
                qualified_unmapped_read = True

            elif len(cigar_splitted) == 2:

                # e.g. 139S61M
                if (cigar_splitted[0][-1] == 'S') and (cigar_splitted[1][-1] == 'M'):
                    cigar_M_pct = int(cigar_splitted[1][:-1]) * 100 / read_len
                    if (ref_pos == 1) and (cigar_M_pct >= 25):
                        qualified_unmapped_read = True

                # e.g. 147M53S
                if (cigar_splitted[0][-1] == 'M') and (cigar_splitted[1][-1] == 'S'):
                    cigar_M_pct = int(cigar_splitted[0][:-1]) * 100 / read_len
                    matched_to_bp = ref_pos + int(cigar_splitted[0][:-1]) - 1
                    if (matched_to_bp == ref_len) and (cigar_M_pct >= 25):
                        qualified_unmapped_read = True

            elif len(cigar_splitted) == 3:

                # e.g. 121M1D66M, 121M1I66M
                if (cigar_splitted[0][-1] == 'M') and (cigar_splitted[2][-1] == 'M'):
                    mismatch_pct = int(cigar_splitted[1][:-1])*100/read_len
                    if mismatch_pct <= 1:
                        qualified_unmapped_read = True

                # e.g. 11S81M3S, not sure, not considered yet
                elif cigar_splitted[1][-1] == 'M':
                    min_mismatch = min(int(cigar_splitted[0][:-1]), int(cigar_splitted[2][:-1]))
                    if min_mismatch <= 3:
                        pass
            elif len(cigar_splitted) == 4:

                # e.g. ['181M', '1D', '8M', '11S'], ['181M', '2D', '8M', '11S']
                if (cigar_splitted[0][-1] == 'M') and (cigar_splitted[2][-1] == 'M') and (cigar_splitted[3][-1] == 'S'):
                    cigar_M_pct = (int(cigar_splitted[0][:-1]) + int(cigar_splitted[2][:-1]))*100/read_len
                    if ((ref_pos + int(cigar_splitted[0][:-1]) + int(cigar_splitted[2][:-1])) == ref_len) and (int(cigar_splitted[1][:-1]) <= 2):
                        if cigar_M_pct >= 25:
                            qualified_unmapped_read = True

                # e.g. ['24S', '95M', '1D', '27M'] and ref_pos = 1
                elif (cigar_splitted[0][-1] == 'S') and (cigar_splitted[1][-1] == 'M') and (cigar_splitted[3][-1] == 'M'):
                    cigar_M_pct = (int(cigar_splitted[1][:-1]) + int(cigar_splitted[3][:-1]))*100/read_len
                    if (ref_pos == 1) and (int(cigar_splitted[2][:-1]) <= 1) and (cigar_M_pct >= 25):
                        qualified_unmapped_read = True

            # add to dict
            if qualified_unmapped_read is True:
                if read_id not in unmapped_paired_reads_to_ctg_dict_by_mapping:
                    unmapped_paired_reads_to_ctg_dict_by_mapping[read_id] = [ref_id_with_prefix]
                else:
                    unmapped_paired_reads_to_ctg_dict_by_mapping[read_id].append(ref_id_with_prefix)


    return unmapped_paired_reads_to_ctg_dict_by_mapping


unmapped_paired_reads_to_ctg_dict_by_mapping = paired_blast_results_to_dict_by_mapping(pwd_samfile_to_mag)

print('unmapped_paired_reads_to_ctg_dict')
print(unmapped_paired_reads_to_ctg_dict)
print(len(unmapped_paired_reads_to_ctg_dict))
print()
print('unmapped_paired_reads_to_ctg_dict_by_mapping')
print(unmapped_paired_reads_to_ctg_dict_by_mapping)
print(len(unmapped_paired_reads_to_ctg_dict_by_mapping))




paired_stats_dict_num = {}
paired_reads_match_profile_handle = open(paired_reads_match_profile, 'w')
paired_reads_match_profile_handle.write('ID\tR1\tR2\n')
for unmapped_read in unmapped_paired_reads_to_ctg_dict:

    unmapped_read_base = '.'.join(unmapped_read.split('.')[:-1])
    unmapped_read_strand = unmapped_read.split('.')[-1]

    unmapped_read_base_r1_matched = []
    unmapped_read_base_r2_matched = []
    if unmapped_read_strand == '1':
        unmapped_read_base_r1_matched = unmapped_paired_reads_to_ctg_dict[unmapped_read]
        if unmapped_read_base in perfectly_mapped_read_singleton_dict:
            unmapped_read_base_r2_matched = perfectly_mapped_read_singleton_dict[unmapped_read_base]['2']
    if unmapped_read_strand == '2':
        unmapped_read_base_r2_matched = unmapped_paired_reads_to_ctg_dict[unmapped_read]
        if unmapped_read_base in perfectly_mapped_read_singleton_dict:
            unmapped_read_base_r1_matched = perfectly_mapped_read_singleton_dict[unmapped_read_base]['1']

    if (unmapped_read_base_r1_matched != []) and (unmapped_read_base_r2_matched != []):
        for r1 in unmapped_read_base_r1_matched:
            for r2 in unmapped_read_base_r2_matched:

                # write out to file
                paired_reads_match_profile_handle.write('%s\t%s\t%s\n' % (unmapped_read_base, r1, r2))

                # store in dict
                paired_key = '_|_'.join(sorted([r1, r2])[::-1])
                if genomic_seq_type == 'mag':
                    paired_key = '___'.join(paired_key.split('___')[:-1])
                if paired_key not in paired_stats_dict_num:
                    paired_stats_dict_num[paired_key] = 1
                else:
                    paired_stats_dict_num[paired_key] += 1

paired_reads_match_profile_handle.close()


'''
cd /srv/scratch/z5039045/MarkerMAG_wd/Kelp/BH_ER_050417_Mira_MarkerMAG_wd/BH_ER_050417_Mira_step_1_wd/BH_ER_050417_refined_bins_db
bowtie2-build -f BH_ER_050417_refined_bins_combined.fa BH_ER_050417_refined_bins_combined --quiet --threads 12
bowtie2 -x BH_ER_050417_refined_bins_combined -U ../unmapped_paired_reads.fasta -S BH_ER_050417_refined_bins_combined.sam -f --local --no-unal --quiet --threads 12


178
205
209
210

'''


import os
import pandas as pd


def sort_csv_by_col(file_in, file_out, col_header):
    df_in        = pd.read_csv(file_in)
    df_in_sorted = df_in.sort_values(by=[col_header], ascending=False)
    df_in_sorted.to_csv(file_out, index=False)


def sep_path_basename_ext(file_in):

    # separate path and file name
    file_path, file_name = os.path.split(file_in)
    if file_path == '':
        file_path = '.'

    # separate file basename and extension
    file_basename, file_extension = os.path.splitext(file_name)

    return file_path, file_basename, file_extension


def blast_results_to_pairwise_16s_iden_dict(blastn_output, align_len_cutoff, cov_cutoff):

    pairwise_iden_dict = {}
    for match in open(blastn_output):
        match_split = match.strip().split('\t')
        query = match_split[0]
        subject = match_split[1]
        iden = float(match_split[2])
        align_len = int(match_split[3])
        query_len = int(match_split[12])
        subject_len = int(match_split[13])
        coverage_q = float(align_len) * 100 / float(query_len)
        coverage_s = float(align_len) * 100 / float(subject_len)

        if (align_len >= align_len_cutoff) and (query != subject) and (coverage_q >= cov_cutoff) and (coverage_s >= cov_cutoff):
            query_to_subject_key = '__|__'.join(sorted([query, subject]))
            if query_to_subject_key not in pairwise_iden_dict:
                pairwise_iden_dict[query_to_subject_key] = iden
            else:
                if iden > pairwise_iden_dict[query_to_subject_key]:
                    pairwise_iden_dict[query_to_subject_key] = iden

    return pairwise_iden_dict


def filter_linkages_iteratively(file_in, sort_by_col_header, pairwise_16s_iden_dict, genomic_seq_depth_dict, marker_gene_depth_dict, min_16s_gnm_multiple, within_genome_16s_divergence_cutoff, min_linkages, min_linkages_for_uniq_linked_16s, file_out):

    # get MarkerGene_to_GenomicSeq_dict
    MarkerGene_to_GenomicSeq_dict = {}
    for each_linkage in open(file_in):
        if not each_linkage.startswith('MarkerGene,GenomicSeq,Number'):
            each_linkage_split = each_linkage.strip().split(',')
            MarkerGene_id = each_linkage_split[0][12:]
            GenomicSeq_id = each_linkage_split[1][12:]
            linkage_num = int(each_linkage_split[2])
            if linkage_num > 1:
                if MarkerGene_id not in MarkerGene_to_GenomicSeq_dict:
                    MarkerGene_to_GenomicSeq_dict[MarkerGene_id] = {GenomicSeq_id}
                else:
                    MarkerGene_to_GenomicSeq_dict[MarkerGene_id].add(GenomicSeq_id)


    file_in_path, file_in_basename, file_in_extension = sep_path_basename_ext(file_in)
    file_in_sorted = '%s/%s_sorted%s' % (file_in_path, file_in_basename, file_in_extension)

    # sort file in
    sort_csv_by_col(file_in, file_in_sorted, sort_by_col_header)

    # fileter linkage
    file_out_handle = open(file_out, 'w')
    MarkerGene_with_assignment = set()
    GenomicSeq_best_marker_dict = {}
    for each_match in open(file_in_sorted):

        if each_match.startswith('MarkerGene,GenomicSeq,Number'):
            file_out_handle.write(each_match)
        else:
            match_split = each_match.strip().split(',')
            MarkerGene = match_split[0][12:]
            GenomicSeq = match_split[1][12:]
            linkage_num = int(match_split[2])

            current_min_linkage = min_linkages_for_uniq_linked_16s
            if MarkerGene in MarkerGene_to_GenomicSeq_dict:
                if len(MarkerGene_to_GenomicSeq_dict[MarkerGene]) > 1:
                    current_min_linkage = min_linkages

            if linkage_num >= current_min_linkage:

                # consider depth
                if min_16s_gnm_multiple > 0:
                    MarkerGene_depth = marker_gene_depth_dict[MarkerGene]
                    GenomicSeq_depth = genomic_seq_depth_dict[GenomicSeq]
                    if (MarkerGene_depth/GenomicSeq_depth) >= min_16s_gnm_multiple:
                        if MarkerGene not in MarkerGene_with_assignment:

                            if GenomicSeq not in GenomicSeq_best_marker_dict:
                                GenomicSeq_best_marker_dict[GenomicSeq] = MarkerGene
                                file_out_handle.write(each_match)
                                MarkerGene_with_assignment.add(MarkerGene)
                            else:
                                current_GenomicSeq_best_marker = GenomicSeq_best_marker_dict[GenomicSeq]
                                key_str = '__|__'.join(sorted([MarkerGene, current_GenomicSeq_best_marker]))

                                iden_with_best_marker = 0
                                if key_str in pairwise_16s_iden_dict:
                                    iden_with_best_marker = pairwise_16s_iden_dict[key_str]

                                if iden_with_best_marker >= within_genome_16s_divergence_cutoff:
                                    file_out_handle.write(each_match)
                                    MarkerGene_with_assignment.add(MarkerGene)
                # ignore depth
                else:
                    if MarkerGene not in MarkerGene_with_assignment:
                        if each_match == 'MarkerGene__DM_m2 matam_MBARC26_subsample_5_143,GenomicSeq__DM,444\n':
                            print(each_match)

                        if GenomicSeq not in GenomicSeq_best_marker_dict:
                            GenomicSeq_best_marker_dict[GenomicSeq] = MarkerGene
                            file_out_handle.write(each_match)
                            MarkerGene_with_assignment.add(MarkerGene)
                        else:



                            current_GenomicSeq_best_marker = GenomicSeq_best_marker_dict[GenomicSeq]
                            key_str = '__|__'.join(sorted([MarkerGene, current_GenomicSeq_best_marker]))

                            iden_with_best_marker = 0
                            if key_str in pairwise_16s_iden_dict:
                                iden_with_best_marker = pairwise_16s_iden_dict[key_str]

                            if each_match == 'MarkerGene__DM_m2 matam_MBARC26_subsample_5_143,GenomicSeq__DM,444\n':
                                print(each_match)
                                print(key_str)
                                print(iden_with_best_marker)

                            if iden_with_best_marker >= within_genome_16s_divergence_cutoff:
                                file_out_handle.write(each_match)
                                MarkerGene_with_assignment.add(MarkerGene)

    file_out_handle.close()

    # remove tmp file
    # os.remove(file_in_sorted)


wd                                  = '/Users/songweizhi/Desktop/000'
file_in                             = '%s/stats_paired.txt' % wd
sort_by_col_header                  = 'Number'
genomic_seq_depth_dict              = {}
marker_gene_depth_dict              = {}
min_16s_gnm_multiple                = 0
within_genome_16s_divergence_cutoff = 98
min_linkages                        = 5
min_linkages_for_uniq_linked_16s    = 3
file_out                            = '%s/stats_paired_filtered.txt' % wd
blast_results_all_vs_all_16s        = '%s/16S_all_vs_all_blastn.tab' % wd
aln16s                              = 500
cov16s                              = 30

pairwise_16s_iden_dict = blast_results_to_pairwise_16s_iden_dict(blast_results_all_vs_all_16s, aln16s, cov16s)
filter_linkages_iteratively(file_in, sort_by_col_header, pairwise_16s_iden_dict, genomic_seq_depth_dict, marker_gene_depth_dict, min_16s_gnm_multiple, within_genome_16s_divergence_cutoff, min_linkages, min_linkages_for_uniq_linked_16s, file_out)

print(pairwise_16s_iden_dict)



def cigar_splitter():
    pass

def check_both_ends_clipping(aaa):
    pass

def cigar_splitter(cifar):
    return []

def get_cigar_stats():
    pass

aa = 0




def check_cigar_quality(cigar_str, mismatch_cutoff, min_M_len, ref_pos, ref_len):

    r2_ctg_ref_cigar_splitted = cigar_splitter(cigar_str)
    qualified_cigar = False

    # check both end clip
    both_end_clp = check_both_ends_clipping(r2_ctg_ref_cigar_splitted)
    if both_end_clp is False:
        # check mismatch
        r2_aligned_len_ctg, r2_aligned_pct_ctg, r2_clipping_len_ctg, r2_clipping_pct_ctg, r2_mismatch_pct_ctg = get_cigar_stats(r2_ctg_ref_cigar_splitted)
        if r2_mismatch_pct_ctg <= mismatch_cutoff:
            # check aligned length
            if r2_aligned_len_ctg >= min_M_len:
                # check if clp in the middle
                clip_in_middle = False
                if ('S' in cigar_str) or ('s' in cigar_str):
                    clip_in_middle = True
                    if (r2_ctg_ref_cigar_splitted[0][-1] in ['S', 's']) and (ref_pos == 1):
                        clip_in_middle = False
                    if (r2_ctg_ref_cigar_splitted[-1][-1] in ['S', 's']):
                        if (ref_pos + r2_aligned_len_ctg - 1) == ref_len:
                            clip_in_middle = False

                if clip_in_middle is False:
                    qualified_cigar = True

    return qualified_cigar






js_header = '''#!/bin/bash
#PBS -l nodes=1:ppn=1
#PBS -l mem=30gb
#PBS -l walltime=11:59:00
#PBS -j oe
#PBS -M weizhi.song@unsw.edu.au
#PBS -m ae

cd /srv/scratch/z5039045/MarkerMAG_wd/Tara_SOC/1_raw_reads
'''

js_folder = '/Users/songweizhi/Desktop/Tara_SOC_js_wget'


for each_sample in ['wget ftp://ftp.sra.ebi.ac.uk/vol1/fastq/ERR598/ERR598945/ERR598945_1.fastq.gz', 'wget ftp://ftp.sra.ebi.ac.uk/vol1/fastq/ERR598/ERR598945/ERR598945_2.fastq.gz', 'wget ftp://ftp.sra.ebi.ac.uk/vol1/fastq/ERR599/ERR599059/ERR599059_1.fastq.gz', 'wget ftp://ftp.sra.ebi.ac.uk/vol1/fastq/ERR599/ERR599059/ERR599059_2.fastq.gz', 'wget ftp://ftp.sra.ebi.ac.uk/vol1/fastq/ERR358/008/ERR3587138/ERR3587138_1.fastq.gz', 'wget ftp://ftp.sra.ebi.ac.uk/vol1/fastq/ERR358/008/ERR3587138/ERR3587138_2.fastq.gz', 'wget ftp://ftp.sra.ebi.ac.uk/vol1/fastq/ERR358/001/ERR3587191/ERR3587191_1.fastq.gz', 'wget ftp://ftp.sra.ebi.ac.uk/vol1/fastq/ERR358/001/ERR3587191/ERR3587191_2.fastq.gz', 'wget ftp://ftp.sra.ebi.ac.uk/vol1/fastq/ERR599/ERR599090/ERR599090_1.fastq.gz', 'wget ftp://ftp.sra.ebi.ac.uk/vol1/fastq/ERR599/ERR599090/ERR599090_2.fastq.gz', 'wget ftp://ftp.sra.ebi.ac.uk/vol1/fastq/ERR599/ERR599176/ERR599176_1.fastq.gz', 'wget ftp://ftp.sra.ebi.ac.uk/vol1/fastq/ERR599/ERR599176/ERR599176_2.fastq.gz', 'wget ftp://ftp.sra.ebi.ac.uk/vol1/fastq/ERR358/009/ERR3587139/ERR3587139_1.fastq.gz', 'wget ftp://ftp.sra.ebi.ac.uk/vol1/fastq/ERR358/009/ERR3587139/ERR3587139_2.fastq.gz', 'wget ftp://ftp.sra.ebi.ac.uk/vol1/fastq/ERR358/002/ERR3587192/ERR3587192_1.fastq.gz', 'wget ftp://ftp.sra.ebi.ac.uk/vol1/fastq/ERR358/002/ERR3587192/ERR3587192_2.fastq.gz', 'wget ftp://ftp.sra.ebi.ac.uk/vol1/fastq/ERR599/ERR599104/ERR599104_1.fastq.gz', 'wget ftp://ftp.sra.ebi.ac.uk/vol1/fastq/ERR599/ERR599104/ERR599104_2.fastq.gz', 'wget ftp://ftp.sra.ebi.ac.uk/vol1/fastq/ERR599/ERR599121/ERR599121_1.fastq.gz', 'wget ftp://ftp.sra.ebi.ac.uk/vol1/fastq/ERR599/ERR599121/ERR599121_2.fastq.gz', 'wget ftp://ftp.sra.ebi.ac.uk/vol1/fastq/ERR358/000/ERR3587140/ERR3587140_1.fastq.gz', 'wget ftp://ftp.sra.ebi.ac.uk/vol1/fastq/ERR358/000/ERR3587140/ERR3587140_2.fastq.gz', 'wget ftp://ftp.sra.ebi.ac.uk/vol1/fastq/ERR358/003/ERR3587193/ERR3587193_1.fastq.gz', 'wget ftp://ftp.sra.ebi.ac.uk/vol1/fastq/ERR358/003/ERR3587193/ERR3587193_2.fastq.gz']:

    sample_id = each_sample.split('/')[-1].split('.')[0]
    pwd_js = '%s/js_%s.sh' % (js_folder, sample_id)
    pwd_js_handle = open(pwd_js, 'w')
    pwd_js_handle.write(js_header)
    pwd_js_handle.write(each_sample)
    pwd_js_handle.close()



import os
import pandas as pd


def sep_path_basename_ext(file_in):

    # separate path and file name
    file_path, file_name = os.path.split(file_in)
    if file_path == '':
        file_path = '.'

    # separate file basename and extension
    file_basename, file_extension = os.path.splitext(file_name)

    return file_path, file_basename, file_extension


def blast_results_to_pairwise_16s_iden_dict(blastn_output, align_len_cutoff, cov_cutoff):

    pairwise_iden_dict = {}
    for match in open(blastn_output):
        match_split = match.strip().split('\t')
        query = match_split[0]
        subject = match_split[1]
        iden = float(match_split[2])
        align_len = int(match_split[3])
        query_len = int(match_split[12])
        subject_len = int(match_split[13])
        coverage_q = float(align_len) * 100 / float(query_len)
        coverage_s = float(align_len) * 100 / float(subject_len)

        if (align_len >= align_len_cutoff) and (query != subject) and (coverage_q >= cov_cutoff) and (coverage_s >= cov_cutoff):
            query_to_subject_key = '__|__'.join(sorted([query, subject]))
            if query_to_subject_key not in pairwise_iden_dict:
                pairwise_iden_dict[query_to_subject_key] = iden
            else:
                if iden > pairwise_iden_dict[query_to_subject_key]:
                    pairwise_iden_dict[query_to_subject_key] = iden

    return pairwise_iden_dict


def sort_csv_by_col(file_in, file_out, col_header):
    df_in        = pd.read_csv(file_in)
    df_in_sorted = df_in.sort_values(by=[col_header], ascending=False)
    df_in_sorted.to_csv(file_out, index=False)


def filter_linkages_iteratively_backup(file_in, sort_by_col_header, pairwise_16s_iden_dict, genomic_seq_depth_dict, marker_gene_depth_dict, min_16s_gnm_multiple, within_genome_16s_divergence_cutoff, min_linkages, min_linkages_for_uniq_linked_16s, file_out):

    # get MarkerGene_to_GenomicSeq_dict
    MarkerGene_to_GenomicSeq_dict = {}
    for each_linkage in open(file_in):
        if not each_linkage.startswith('MarkerGene,GenomicSeq,Number'):
            each_linkage_split = each_linkage.strip().split(',')
            MarkerGene_id = each_linkage_split[0][12:]
            GenomicSeq_id = each_linkage_split[1][12:]
            linkage_num = int(each_linkage_split[2])
            if linkage_num > 1:
                if MarkerGene_id not in MarkerGene_to_GenomicSeq_dict:
                    MarkerGene_to_GenomicSeq_dict[MarkerGene_id] = {GenomicSeq_id}
                else:
                    MarkerGene_to_GenomicSeq_dict[MarkerGene_id].add(GenomicSeq_id)


    file_in_path, file_in_basename, file_in_extension = sep_path_basename_ext(file_in)
    file_in_sorted = '%s/%s_sorted%s' % (file_in_path, file_in_basename, file_in_extension)

    # sort file in
    sort_csv_by_col(file_in, file_in_sorted, sort_by_col_header)

    # fileter linkage
    gnm_max_link_num_dict = {}
    file_out_handle = open(file_out, 'w')
    MarkerGene_with_assignment = set()
    GenomicSeq_best_marker_dict = {}
    for each_match in open(file_in_sorted):
        if each_match.startswith('MarkerGene,GenomicSeq,Number'):
            file_out_handle.write(each_match)
        else:
            match_split = each_match.strip().split(',')
            MarkerGene = match_split[0][12:]
            GenomicSeq = match_split[1][12:]
            linkage_num = int(match_split[2])

            current_min_linkage = min_linkages_for_uniq_linked_16s
            if MarkerGene in MarkerGene_to_GenomicSeq_dict:
                if len(MarkerGene_to_GenomicSeq_dict[MarkerGene]) > 1:
                    current_min_linkage = min_linkages

            if linkage_num >= current_min_linkage:

                # consider depth
                if min_16s_gnm_multiple > 0:
                    MarkerGene_depth = marker_gene_depth_dict[MarkerGene]
                    GenomicSeq_depth = genomic_seq_depth_dict[GenomicSeq]
                    if (MarkerGene_depth/GenomicSeq_depth) >= min_16s_gnm_multiple:
                        if MarkerGene not in MarkerGene_with_assignment:

                            if GenomicSeq not in GenomicSeq_best_marker_dict:
                                GenomicSeq_best_marker_dict[GenomicSeq] = MarkerGene
                                file_out_handle.write(each_match)
                                MarkerGene_with_assignment.add(MarkerGene)
                            else:
                                current_GenomicSeq_best_marker = GenomicSeq_best_marker_dict[GenomicSeq]
                                key_str = '__|__'.join(sorted([MarkerGene, current_GenomicSeq_best_marker]))

                                iden_with_best_marker = 0
                                if key_str in pairwise_16s_iden_dict:
                                    iden_with_best_marker = pairwise_16s_iden_dict[key_str]

                                if iden_with_best_marker >= within_genome_16s_divergence_cutoff:
                                    file_out_handle.write(each_match)
                                    MarkerGene_with_assignment.add(MarkerGene)
                # ignore depth
                else:

                    if MarkerGene not in MarkerGene_with_assignment:
                        if GenomicSeq not in GenomicSeq_best_marker_dict:
                            GenomicSeq_best_marker_dict[GenomicSeq] = MarkerGene
                            file_out_handle.write(each_match)
                            MarkerGene_with_assignment.add(MarkerGene)

                            # store max link num
                            gnm_max_link_num_dict[GenomicSeq] = linkage_num

                        else:
                            current_GenomicSeq_best_marker = GenomicSeq_best_marker_dict[GenomicSeq]
                            key_str = '__|__'.join(sorted([MarkerGene, current_GenomicSeq_best_marker]))

                            iden_with_best_marker = 0
                            if key_str in pairwise_16s_iden_dict:
                                iden_with_best_marker = pairwise_16s_iden_dict[key_str]

                            if iden_with_best_marker >= within_genome_16s_divergence_cutoff:
                                file_out_handle.write(each_match)
                                MarkerGene_with_assignment.add(MarkerGene)
    file_out_handle.close()


def filter_linkages_iteratively(file_in, sort_by_col_header, pairwise_16s_iden_dict, genomic_seq_depth_dict,
                                marker_gene_depth_dict, min_16s_gnm_multiple, within_genome_16s_divergence_cutoff,
                                min_linkages, min_linkages_for_uniq_linked_16s, file_out):

    # get MarkerGene_to_GenomicSeq_dict
    MarkerGene_to_GenomicSeq_dict = {}
    for each_linkage in open(file_in):
        if not each_linkage.startswith('MarkerGene,GenomicSeq,Number'):
            each_linkage_split = each_linkage.strip().split(',')
            MarkerGene_id = each_linkage_split[0][12:]
            GenomicSeq_id = each_linkage_split[1][12:]
            linkage_num = int(each_linkage_split[2])
            if linkage_num > 1:
                if MarkerGene_id not in MarkerGene_to_GenomicSeq_dict:
                    MarkerGene_to_GenomicSeq_dict[MarkerGene_id] = {GenomicSeq_id}
                else:
                    MarkerGene_to_GenomicSeq_dict[MarkerGene_id].add(GenomicSeq_id)

    file_in_path, file_in_basename, file_in_extension = sep_path_basename_ext(file_in)
    file_in_sorted = '%s/%s_sorted%s' % (file_in_path, file_in_basename, file_in_extension)

    # sort file in
    sort_csv_by_col(file_in, file_in_sorted, sort_by_col_header)

    # fileter linkage
    gnm_max_link_num_dict = {}
    file_out_handle = open(file_out, 'w')
    MarkerGene_with_assignment = set()
    GenomicSeq_best_marker_dict = {}
    for each_match in open(file_in_sorted):
        if each_match.startswith('MarkerGene,GenomicSeq,Number'):
            file_out_handle.write(each_match)
        else:
            match_split = each_match.strip().split(',')
            MarkerGene = match_split[0][12:]
            GenomicSeq = match_split[1][12:]
            linkage_num = int(match_split[2])

            current_min_linkage = min_linkages_for_uniq_linked_16s
            if MarkerGene in MarkerGene_to_GenomicSeq_dict:
                if len(MarkerGene_to_GenomicSeq_dict[MarkerGene]) > 1:
                    current_min_linkage = min_linkages

            if linkage_num >= current_min_linkage:

                if MarkerGene not in MarkerGene_with_assignment:


                    if GenomicSeq not in GenomicSeq_best_marker_dict:
                        GenomicSeq_best_marker_dict[GenomicSeq] = MarkerGene
                        gnm_max_link_num_dict[GenomicSeq] = linkage_num
                        file_out_handle.write(each_match)
                        MarkerGene_with_assignment.add(MarkerGene)
                    else:
                        # get identity with best marker
                        current_GenomicSeq_best_marker = GenomicSeq_best_marker_dict[GenomicSeq]
                        key_str = '__|__'.join(sorted([MarkerGene, current_GenomicSeq_best_marker]))
                        iden_with_best_marker = pairwise_16s_iden_dict.get(key_str, 0)

                        if iden_with_best_marker >= within_genome_16s_divergence_cutoff:

                            gnm_max_link_num = gnm_max_link_num_dict[GenomicSeq]
                            if (linkage_num*100/gnm_max_link_num) >= 80:
                                file_out_handle.write(each_match)
                                MarkerGene_with_assignment.add(MarkerGene)
                            else:
                                MarkerGene_with_assignment.add(MarkerGene)

    file_out_handle.close()




'''
	Round	|	Link	Yes	NA	No	Accuracy	|	MAG	Yes	NA	No	Y/N	Recovery	Accuracy
raw	Rd_1	|	140	119	9	12	119/131(90.84)	|	32	29	0	0	3	29/97(29.9)	    29/32(90.62)
33	Rd_1	|	86	82	2	2	82/84(97.62)	|	32	30	0	0	2	30/97(30.93)	30/32(93.75)
50	Rd_1	|	73	69	1	3	69/72(95.83)	|	34	31	0	2	1	31/97(31.96)	31/34(91.18)
75  Rd_1	|	55	53	1	1	53/54(98.15)	|	34	33	0	1	0	33/97(34.02)	33/34(97.06)
85	Rd_1	|	46	44	1	1	44/45(97.78)	|	34	33	0	1	0	33/97(34.02)	33/34(97.06)
90	Rd_1	|	43	41	1	1	41/42(97.62)	|	34	33	0	1	0	33/97(34.02)	33/34(97.06)


33  Rd_1	|	83	79	2	2	79/81(97.53)	|	32	30	0	0	2	30/97(30.93)	30/32(93.75)
50  Rd_1	|	71	69	1	1	69/70(98.57)	|	32	31	0	0	1	31/97(31.96)	31/32(96.88)
75  Rd_1	|	51	50	1	0	50/50(100.0)	|	32	32	0	0	0	32/97(32.99)	32/32(100.0)
80	Rd_1	|	50	49	1	0	49/49(100.0)	|	32	32	0	0	0	32/97(32.99)	32/32(100.0)
85	Rd_1	|	44	43	1	0	43/43(100.0)	|	32	32	0	0	0	32/97(32.99)	32/32(100.0)
90  Rd_1	|	41	40	1	0	40/40(100.0)	|	32	32	0	0	0	32/97(32.99)	32/32(100.0)

'''

step_1_wd                       = '/Users/songweizhi/Desktop/step_1_wd'
link_stats_combined             = '%s/stats_combined.txt'               % step_1_wd
link_stats_combined_filtered_s1 = '%s/stats_combined_filtered.txt'      % step_1_wd
blast_results_all_vs_all_16s    = '%s/16S_all_vs_all_blastn.tab' % step_1_wd
min_aln_16s                     = 500
min_cov_16s                     = 30
mean_depth_dict_gnm             = {}
mean_depth_dict_16s             = {}
min_16s_gnm_multiple            = 0
min_iden_16s                    = 98
min_link_num_rd1                = 10
min_uniq_link_num_rd1           = 7

combined_linkage_file = '%s/Test_identified_linkages_genome_level.txt' % step_1_wd

pairwise_16s_iden_dict = blast_results_to_pairwise_16s_iden_dict(blast_results_all_vs_all_16s, min_aln_16s, min_cov_16s)


filter_linkages_iteratively(link_stats_combined, 'Number', pairwise_16s_iden_dict,
                            mean_depth_dict_gnm, mean_depth_dict_16s, min_16s_gnm_multiple,
                            min_iden_16s, min_link_num_rd1,
                            min_uniq_link_num_rd1, link_stats_combined_filtered_s1)

combined_linkage_file_handle = open(combined_linkage_file, 'w')
combined_linkage_file_handle.write('MarkerGene\tGenomicSeq\tLinkage\tStep\n')
for step_1_link in open(link_stats_combined_filtered_s1):
    if not step_1_link.startswith('MarkerGene,GenomicSeq,Number'):
        marker_id = step_1_link.strip().split(',')[0][12:]
        genome_id = step_1_link.strip().split(',')[1][12:]
        link_num = step_1_link.strip().split(',')[2]
        combined_linkage_file_handle.write('%s\t%s\t%s\tS1\n' % (marker_id, genome_id, link_num))
combined_linkage_file_handle.close()

os.system('python3 /Users/songweizhi/PycharmProjects/MarkerMAG/script_backup/GI/assess_linkages_GI.py')




import os
import numpy as np
from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord


def cigar_splitter(cigar):
    # get the position of letters
    letter_pos_list = []
    n = 0
    for each_element in cigar:
        if (each_element.isalpha() is True) or (each_element == '='):
            letter_pos_list.append(n)
        n += 1

    # split cigar
    index = 0
    cigar_splitted = []
    while index <= len(letter_pos_list) - 1:
        if index == 0:
            cigar_splitted.append(cigar[:(letter_pos_list[index] + 1)])
        else:
            cigar_splitted.append(cigar[(letter_pos_list[index - 1] + 1):(letter_pos_list[index] + 1)])
        index += 1

    return cigar_splitted


def get_cigar_stats(cigar_splitted):
    # aligned_len: M I X =
    # clipping_len: S
    # mismatch_len: X I D
    # mismatch_pct = mismatch_len / aligned_len
    # aligned_pct  = aligned_len  / (aligned_len + clipping_len)
    # clipping_pct = clipping_len / (aligned_len + clipping_len)

    aligned_len = 0
    clipping_len = 0
    mismatch_len = 0
    for each_part in cigar_splitted:
        each_part_len = int(each_part[:-1])
        each_part_cate = each_part[-1]

        # get aligned_len
        if each_part_cate in {'M', 'm', 'I', 'i', 'X', 'x', '='}:
            aligned_len += each_part_len

        # get clipping_len
        if each_part_cate in ['S', 's']:
            clipping_len += each_part_len

        # get mismatch_len
        if each_part_cate in {'I', 'i', 'X', 'x', 'D', 'd'}:
            mismatch_len += each_part_len

    aligned_pct = float("{0:.2f}".format(aligned_len * 100 / (aligned_len + clipping_len)))
    clipping_pct = float("{0:.2f}".format(clipping_len * 100 / (aligned_len + clipping_len)))
    mismatch_pct = float("{0:.2f}".format(mismatch_len * 100 / (aligned_len)))

    return aligned_len, aligned_pct, clipping_len, clipping_pct, mismatch_pct


def check_both_ends_clipping(cigar_splitted):
    both_ends_clipping = False
    if len(cigar_splitted) >= 3:
        if (cigar_splitted[0][-1] in ['S', 's']) and (cigar_splitted[-1][-1] in ['S', 's']):
            both_ends_clipping = True

    return both_ends_clipping


class MappingRecord:

    def __init__(self):

        self.r1_16s_ref_dict = dict()
        self.r2_16s_ref_dict = dict()

        self.r1_16s_refs_lowest_mismatch = None
        self.r2_16s_refs_lowest_mismatch = None

        self.r1_16s_refs_no_ignored = dict()
        self.r2_16s_refs_no_ignored = dict()
        self.shared_16s_refs_no_ignored = dict()

        self.r1_ctg_ref_dict = dict()
        self.r2_ctg_ref_dict = dict()
        self.r1_ctg_refs_no_ignored = dict()
        self.r2_ctg_refs_no_ignored = dict()
        self.shared_ctg_refs_no_ignored = dict()

        self.r1_longest_clp_cigar = ''
        self.r1_longest_clp_falg = ''
        self.r2_longest_clp_cigar = ''
        self.r2_longest_clp_falg = ''

        self.qualified_reads = False
        self.matched_to_ctg = False
        self.both_mapped_to_16s = False
        self.consider_r1_unmapped_mate = False
        self.consider_r1_clipping_part = False
        self.consider_r2_unmapped_mate = False
        self.consider_r2_clipping_part = False

        self.consider_round_2 = False

        self.r1_filtered_refs = set()
        self.r2_filtered_refs = set()
        self.unmapped_r1_refs = set()
        self.unmapped_r2_refs = set()
        self.clipping_r1_refs = set()
        self.clipping_r2_refs = set()
        self.unmapped_r1_refs_with_pos = set()
        self.unmapped_r2_refs_with_pos = set()
        self.clipping_r1_refs_with_pos = set()
        self.clipping_r2_refs_with_pos = set()


def sam_flag_to_rc(flag_value):
    read_rced = 'na'
    if flag_value != '':
        binary_flag = "{0:b}".format(int(flag_value))
        binary_flag_len = len(str(binary_flag))
        binary_flag_polished = '0' * (12 - binary_flag_len) + str(binary_flag)

        if binary_flag_polished[7] == '0':
            read_rced = False
        if binary_flag_polished[7] == '1':
            read_rced = True

    return read_rced


def get_rc(seq_in):
    seq_in_rc = str(SeqRecord(Seq(seq_in)).reverse_complement().seq)
    return seq_in_rc


def get_max_clp_and_index(r1_cigar_list, r2_cigar_list):
    r1_cigar_list_split = [cigar_splitter(i) for i in r1_cigar_list]
    r2_cigar_list_split = [cigar_splitter(i) for i in r2_cigar_list]

    r1_cigar_list_split_only_clp = []
    for each_r1_cigar_split in r1_cigar_list_split:
        clp_len_l = 0
        if each_r1_cigar_split[0][-1] in ['S', 's']:
            clp_len_l = int(each_r1_cigar_split[0][:-1])
        clp_len_r = 0
        if each_r1_cigar_split[-1][-1] in ['S', 's']:
            clp_len_r = int(each_r1_cigar_split[-1][:-1])
        r1_cigar_list_split_only_clp.append([clp_len_l, clp_len_r])

    r2_cigar_list_split_only_clp = []
    for each_r2_cigar_split in r2_cigar_list_split:
        clp_len_l = 0
        if each_r2_cigar_split[0][-1] in ['S', 's']:
            clp_len_l = int(each_r2_cigar_split[0][:-1])
        clp_len_r = 0
        if each_r2_cigar_split[-1][-1] in ['S', 's']:
            clp_len_r = int(each_r2_cigar_split[-1][:-1])
        r2_cigar_list_split_only_clp.append([clp_len_l, clp_len_r])

    cigar_list_split_only_clp_r1_r2 = [r1_cigar_list_split_only_clp, r2_cigar_list_split_only_clp]

    max_value = 0
    max_value_index = ''
    for num_list_1 in cigar_list_split_only_clp_r1_r2[0]:
        if num_list_1[0] > max_value:
            max_value = num_list_1[0]
            max_value_index = 'r1_l'
        if num_list_1[1] > max_value:
            max_value = num_list_1[1]
            max_value_index = 'r1_r'
    for num_list_2 in cigar_list_split_only_clp_r1_r2[1]:
        if num_list_2[0] > max_value:
            max_value = num_list_2[0]
            max_value_index = 'r2_l'
        if num_list_2[1] > max_value:
            max_value = num_list_2[1]
            max_value_index = 'r2_r'

    # get the best cigar
    best_cigar = ''
    if max_value_index == 'r1_l':
        if best_cigar == '':
            for each_cigar in r1_cigar_list:
                if (each_cigar.startswith('%sS' % max_value)) or (each_cigar.startswith('%ss' % max_value)):
                    best_cigar = each_cigar
    elif max_value_index == 'r1_r':
        if best_cigar == '':
            for each_cigar in r1_cigar_list:
                if (each_cigar.endswith('%sS' % max_value)) or (each_cigar.endswith('%ss' % max_value)):
                    best_cigar = each_cigar
    elif max_value_index == 'r2_l':
        if best_cigar == '':
            for each_cigar in r2_cigar_list:
                if (each_cigar.startswith('%sS' % max_value)) or (each_cigar.startswith('%ss' % max_value)):
                    best_cigar = each_cigar
    elif max_value_index == 'r2_r':
        if best_cigar == '':
            for each_cigar in r2_cigar_list:
                if (each_cigar.endswith('%sS' % max_value)) or (each_cigar.endswith('%ss' % max_value)):
                    best_cigar = each_cigar

    return best_cigar, max_value, max_value_index


def check_cigar_all_clp(cigar_list):
    cigar_all_clp = True
    for each_cigar in cigar_list:
        if ('S' not in each_cigar) and ('s' not in each_cigar):
            cigar_all_clp = False

    return cigar_all_clp


def get_min_max_cigar_S(cigar_list):

    # get cigar_S_set
    cigar_S_set = set()
    for each_cigar in cigar_list:
        if ('S' not in each_cigar) and ('s' not in each_cigar):
            cigar_S_set.add(0)
        else:
            cigar_splitted = cigar_splitter(each_cigar)

            if cigar_splitted[0][-1] in ['S', 's']:
                cigar_S_set.add(int(cigar_splitted[0][:-1]))

            if cigar_splitted[-1][-1] in ['S', 's']:
                cigar_S_set.add(int(cigar_splitted[1][:-1]))

    # get min_cigar_S and max_cigar_S
    min_cigar_S = 0
    max_cigar_S = 0
    if len(cigar_S_set) > 0:
        min_cigar_S = min(cigar_S_set)
        max_cigar_S = max(cigar_S_set)

    return min_cigar_S, max_cigar_S


def get_min_mismatch_from_cigar_list(r1_ref_cigar_set, min_M_len):

    mismatch_set_all_cigar = set()
    mismatch_set_long_M_cigars = set()
    for each_cigar in r1_ref_cigar_set:
        aligned_len, aligned_pct, clipping_len, clipping_pct, mismatch_pct = get_cigar_stats(cigar_splitter(each_cigar))
        mismatch_set_all_cigar.add(mismatch_pct)
        if aligned_len >= min_M_len:
            mismatch_set_long_M_cigars.add(mismatch_pct)

    min_mismatch = 'NA'
    if len(mismatch_set_all_cigar) > 0:
        min_mismatch = min(mismatch_set_all_cigar)
        if len(mismatch_set_long_M_cigars) > 0:
            min_mismatch = min(mismatch_set_long_M_cigars)

    return min_mismatch


########################################################################################################################
'''
------------------------------------------------------------------------------------------------------------

# cami_hc_SILVA138_id99_50_subsample_100_3569
cd /srv/scratch/z5039045/MarkerMAG_wd/CAMI1/3_High_Complexity/hc_0512_tmp_MarkerMAG_wd_up/hc_0512_tmp_step_1_wd
grep -E '@|cami_hc_SILVA138_id99_50_subsample_100_3569' /srv/scratch/z5039045/MarkerMAG_wd/CAMI1/3_High_Complexity/hc_0511_very_sensitive_MarkerMAG_wd/hc_0511_very_sensitive_step_1_wd/hc_0511_very_sensitive_input_reads_to_16S_reformatted.sam > input_reads_to_16S_subset.sam

# cami_hc_14___NODE_13073_length_6967_cov_0.639487
cd /srv/scratch/z5039045/MarkerMAG_wd/CAMI1/3_High_Complexity/hc_0512_tmp_MarkerMAG_wd_up/hc_0512_tmp_step_1_wd
grep -E '@|cami_hc_14___C___NODE_13073_length_6967_cov_0.639487' rd1_extracted_to_gnm_reformatted.sam > rd1_extracted_to_gnm_subset.sam

------------------------------------------------------------------------------------------------------------

'''
# do not remove both ends clip alignments, important !!!!!!

wd                                              = '/Users/songweizhi/Desktop/consider_all'
input_reads_to_16s_sam_best_match               = '%s/input_reads_to_16S_subset.sam'            % wd
rd1_extracted_to_gnm_sam_reformatted_best_match = '%s/rd1_extracted_to_gnm_subset.sam'          % wd

min_M_len_16s                                   = 45
min_M_len_ctg                                   = 45
min_M_pct                                       = 35
mismatch_cutoff                                 = 2
read_to_marker_connector                        = '___r___'
marker_to_ctg_gnm_Key_connector                 = '___M___'
gnm_to_ctg_connector                            = '___C___'
report_interval                                 = 10000

link_stats_combined                             = '%s/hc_0509_stats_combined.txt'               % wd
rd1_clp_pct_diff_txt                            = '%s/rd1_clp_pct_diff.txt'                     % wd
rd1_clp_pct_diff_txt_to_ignore                  = '%s/rd1_clp_pct_diff_to_ignore.txt'           % wd
linking_reads_rd1                               = '%s/linking_reads_rd1.txt'                 % wd

# on katana
# input_reads_to_16s_sam_best_match               = '/srv/scratch/z5039045/MarkerMAG_wd/CAMI1/3_High_Complexity/hc_0511_very_sensitive_MarkerMAG_wd/hc_0511_very_sensitive_step_1_wd/hc_0511_very_sensitive_input_reads_to_16S_reformatted.sam'
# rd1_extracted_to_gnm_sam_reformatted_best_match = '/srv/scratch/z5039045/MarkerMAG_wd/CAMI1/3_High_Complexity/hc_0512_tmp_MarkerMAG_wd_up/hc_0512_tmp_step_1_wd/rd1_extracted_to_gnm_reformatted.sam'
# link_stats_combined                             = 'linkage_stats_combined.txt'
# rd1_clp_pct_diff_txt                            = 'rd1_clp_pct_diff.txt'


########################################################################################################################

marker_len_dict = {}
MappingRecord_dict = {}
for each_read in open(input_reads_to_16s_sam_best_match):
    each_read_split = each_read.strip().split('\t')

    if each_read.startswith('@'):
        marker_id = ''
        marker_len = 0
        for each_element in each_read_split:
            if each_element.startswith('SN:'):
                marker_id = each_element[3:]
            if each_element.startswith('LN:'):
                marker_len = int(each_element[3:])
        marker_len_dict[marker_id] = marker_len
    else:
        cigar = each_read_split[5]
        if cigar != '*':
            read_id = each_read_split[0]
            read_id_base = '.'.join(read_id.split('.')[:-1])
            read_strand = read_id.split('.')[-1]
            ref_id = each_read_split[2]
            ref_pos = int(each_read_split[3])

            if read_id_base not in MappingRecord_dict:
                MappingRecord_dict[read_id_base] = MappingRecord()

            if read_strand == '1':
                if ref_id not in MappingRecord_dict[read_id_base].r1_16s_ref_dict:
                    MappingRecord_dict[read_id_base].r1_16s_ref_dict[ref_id] = {ref_pos:cigar}
                else:
                    MappingRecord_dict[read_id_base].r1_16s_ref_dict[ref_id][ref_pos] = cigar

            if read_strand == '2':
                if ref_id not in MappingRecord_dict[read_id_base].r2_16s_ref_dict:
                    MappingRecord_dict[read_id_base].r2_16s_ref_dict[ref_id] = {ref_pos:cigar}
                else:
                    MappingRecord_dict[read_id_base].r2_16s_ref_dict[ref_id][ref_pos] = cigar


##################################################### parse MappingRecord_dict ####################################################

processed_num = 0
to_extract_read_base = set()
for each_mp in MappingRecord_dict:

    current_mp_record = MappingRecord_dict[each_mp]
    refs_to_ignore = set()

    ########## get lowest mismatch for r1/r2 16s refs ##########

    # get r1_ref_cigar_set
    r1_ref_cigar_set = set()
    for each_pos_dict in current_mp_record.r1_16s_ref_dict.values():
        each_pos_dict_values = {each_pos_dict[i] for i in each_pos_dict}
        r1_ref_cigar_set.update(each_pos_dict_values)

    # get r2_ref_cigar_set
    r2_ref_cigar_set = set()
    for each_pos_dict in current_mp_record.r2_16s_ref_dict.values():
        each_pos_dict_values = {each_pos_dict[i] for i in each_pos_dict}
        r2_ref_cigar_set.update(each_pos_dict_values)

    r1_ref_min_mismatch = get_min_mismatch_from_cigar_list(r1_ref_cigar_set, min_M_len_16s)
    r2_ref_min_mismatch = get_min_mismatch_from_cigar_list(r2_ref_cigar_set, min_M_len_16s)
    MappingRecord_dict[each_mp].r1_16s_refs_lowest_mismatch = r1_ref_min_mismatch
    MappingRecord_dict[each_mp].r2_16s_refs_lowest_mismatch = r2_ref_min_mismatch

    ########## filter r1 16s refs ##########

    r1_16s_refs_passed_qc = {}
    for r1_16s_ref in current_mp_record.r1_16s_ref_dict:
        r1_matched_pos_dict = current_mp_record.r1_16s_ref_dict[r1_16s_ref]

        # one read need to mapped to one 16S only for one time
        if len(r1_matched_pos_dict) > 1:
            refs_to_ignore.add(r1_16s_ref)
        else:
            r1_16s_ref_pos = list(r1_matched_pos_dict.keys())[0]
            r1_16s_ref_cigar = r1_matched_pos_dict[r1_16s_ref_pos]
            r1_16s_ref_cigar_splitted = cigar_splitter(r1_16s_ref_cigar)

            # check both end clip
            both_end_clp = check_both_ends_clipping(r1_16s_ref_cigar_splitted)
            if both_end_clp is True:
                refs_to_ignore.add(r1_16s_ref)
            else:
                # check mismatch
                r1_aligned_len, r1_aligned_pct, r1_clipping_len, r1_clipping_pct, r1_mismatch_pct = get_cigar_stats(
                    r1_16s_ref_cigar_splitted)
                if r1_ref_min_mismatch == 'NA':
                    refs_to_ignore.add(r1_16s_ref)
                elif (r1_mismatch_pct > r1_ref_min_mismatch) or (r1_mismatch_pct > mismatch_cutoff):
                    refs_to_ignore.add(r1_16s_ref)
                else:
                    # check aligned length
                    if r1_aligned_len < min_M_len_16s:
                        refs_to_ignore.add(r1_16s_ref)
                    else:
                        # check if clp in the middle
                        clip_in_middle = False
                        if ('S' in r1_16s_ref_cigar) or ('s' in r1_16s_ref_cigar):
                            clip_in_middle = True
                            if (r1_16s_ref_cigar_splitted[0][-1] in ['S', 's']) and (r1_16s_ref_pos == 1):
                                clip_in_middle = False
                            if (r1_16s_ref_cigar_splitted[-1][-1] in ['S', 's']):
                                if (r1_16s_ref_pos + r1_aligned_len - 1) == marker_len_dict[r1_16s_ref]:
                                    clip_in_middle = False

                        # exclude the ref if clp in the middle is True
                        if clip_in_middle is True:
                            refs_to_ignore.add(r1_16s_ref)
                        else:
                            r1_16s_refs_passed_qc[r1_16s_ref] = [r1_16s_ref_cigar]

    ########## filter r2 16s refs ##########

    r2_16s_refs_passed_qc = {}
    for r2_16s_ref in current_mp_record.r2_16s_ref_dict:
        r2_matched_pos_dict = current_mp_record.r2_16s_ref_dict[r2_16s_ref]

        # one read need to mapped to one 16S only once
        if len(r2_matched_pos_dict) > 1:
            refs_to_ignore.add(r2_16s_ref)
        else:
            r2_16s_ref_pos = list(r2_matched_pos_dict.keys())[0]
            r2_16s_ref_cigar = r2_matched_pos_dict[r2_16s_ref_pos]
            r2_16s_ref_cigar_splitted = cigar_splitter(r2_16s_ref_cigar)

            # check both end clip
            both_end_clp = check_both_ends_clipping(r2_16s_ref_cigar_splitted)
            if both_end_clp is True:
                refs_to_ignore.add(r2_16s_ref)
            else:
                # check mismatch
                r2_aligned_len, r2_aligned_pct, r2_clipping_len, r2_clipping_pct, r2_mismatch_pct = get_cigar_stats(
                    r2_16s_ref_cigar_splitted)
                if r2_ref_min_mismatch == 'NA':
                    refs_to_ignore.add(r2_16s_ref)
                elif (r2_mismatch_pct > r2_ref_min_mismatch) or (r2_mismatch_pct > mismatch_cutoff):
                    refs_to_ignore.add(r2_16s_ref)
                else:
                    # check aligned length
                    if r2_aligned_len < min_M_len_16s:
                        refs_to_ignore.add(r2_16s_ref)
                    else:
                        # check if clp in the middle
                        clip_in_middle = False
                        if ('S' in r2_16s_ref_cigar) or ('s' in r2_16s_ref_cigar):
                            clip_in_middle = True
                            if (r2_16s_ref_cigar_splitted[0][-1] in ['S', 's']) and (r2_16s_ref_pos == 1):
                                clip_in_middle = False
                            if (r2_16s_ref_cigar_splitted[-1][-1] in ['S', 's']):
                                if (r2_16s_ref_pos + r2_aligned_len - 1) == marker_len_dict[r2_16s_ref]:
                                    clip_in_middle = False

                        # exclude the ref if clp in the middle is True
                        if clip_in_middle is True:
                            refs_to_ignore.add(r2_16s_ref)
                        else:
                            r2_16s_refs_passed_qc[r2_16s_ref] = [r2_16s_ref_cigar]

    ####################################################################################################

    r1_16s_refs_no_ignored = {key: value for key, value in r1_16s_refs_passed_qc.items() if key not in refs_to_ignore}
    r2_16s_refs_no_ignored = {key: value for key, value in r2_16s_refs_passed_qc.items() if key not in refs_to_ignore}

    # no mate has no_ignored alignments
    if (len(r1_16s_refs_no_ignored) == 0) and (len(r2_16s_refs_no_ignored) == 0):
        pass

    # only r1 has no_ignored alignments
    elif (len(r1_16s_refs_no_ignored) > 0) and (len(r2_16s_refs_no_ignored) == 0):

        MappingRecord_dict[each_mp].qualified_reads = True
        MappingRecord_dict[each_mp].consider_r1_unmapped_mate = True
        MappingRecord_dict[each_mp].r1_16s_refs_no_ignored = r1_16s_refs_no_ignored
        to_extract_read_base.add(each_mp)

        # if the shortest S longer than min_M_ctg_len, also consider as clipping mapped
        r1_16s_refs_no_ignored_cigar_list = list(r1_16s_refs_no_ignored.values())
        r1_min_cigar_S, r1_max_cigar_S = get_min_max_cigar_S(r1_16s_refs_no_ignored_cigar_list)
        if r1_min_cigar_S >= min_M_len_ctg:
            MappingRecord_dict[each_mp].consider_r1_clipping_part = True
            to_extract_read_base.add(each_mp)

    # only r2 has no_ignored alignments
    elif (len(r1_16s_refs_no_ignored) == 0) and (len(r2_16s_refs_no_ignored) > 0):

        MappingRecord_dict[each_mp].qualified_reads = True
        MappingRecord_dict[each_mp].consider_r2_unmapped_mate = True
        MappingRecord_dict[each_mp].r2_16s_refs_no_ignored = r2_16s_refs_no_ignored
        to_extract_read_base.add(each_mp)

        # if the shortest S longer than min_M_ctg_len, also consider as clipping mapped
        r2_16s_refs_no_ignored_cigar_list = list(r2_16s_refs_no_ignored.values())
        r2_min_cigar_S, r2_max_cigar_S = get_min_max_cigar_S(r2_16s_refs_no_ignored_cigar_list)
        if r2_min_cigar_S >= min_M_len_ctg:
            MappingRecord_dict[each_mp].consider_r2_clipping_part = True
            to_extract_read_base.add(each_mp)

    # both r1 and r2 have no_ignored alignments
    else:
        shared_16s_ref_dict = {key: [r1_16s_refs_no_ignored[key][0], r2_16s_refs_no_ignored[key][0]] for key in set(r1_16s_refs_no_ignored).intersection(set(r2_16s_refs_no_ignored))}
        if len(shared_16s_ref_dict) > 0:
            MappingRecord_dict[each_mp].qualified_reads = True
            MappingRecord_dict[each_mp].both_mapped_to_16s = True
            MappingRecord_dict[each_mp].shared_16s_refs_no_ignored = shared_16s_ref_dict
            to_extract_read_base.add(each_mp)

    processed_num += 1
    if (processed_num % report_interval == 0):
        processed_pct = processed_num*100/len(MappingRecord_dict)
        processed_pct = float("{0:.2f}".format(processed_pct))
        print('Processed %sk (%s%s)' % (int(processed_num/1000), processed_pct, '%'))


# remove unqualified mapping record from dict
for each_mp in MappingRecord_dict.copy():
    if MappingRecord_dict[each_mp].qualified_reads is False:
        MappingRecord_dict.pop(each_mp)


######################################### read mapping results of rd1 extracted mates into mp dict  #########################################

ctg_len_dict = {}
for each_read in open(rd1_extracted_to_gnm_sam_reformatted_best_match):
    each_read_split = each_read.strip().split('\t')

    if each_read.startswith('@'):
        ctg_id = ''
        ctg_len = 0
        for each_element in each_read_split:
            if each_element.startswith('SN:'):
                ctg_id = each_element[3:]
            if each_element.startswith('LN:'):
                ctg_len = int(each_element[3:])
        ctg_len_dict[ctg_id] = ctg_len
    else:
        cigar = each_read_split[5]
        if cigar != '*':
            read_id = each_read_split[0]
            read_id_base = '.'.join(read_id.split('.')[:-1])
            read_strand = read_id.split('.')[-1]
            ref_id = each_read_split[2]
            ref_pos = int(each_read_split[3])

            if read_id_base not in MappingRecord_dict:
                MappingRecord_dict[read_id_base] = MappingRecord()

            if read_strand == '1':
                if ref_id not in MappingRecord_dict[read_id_base].r1_ctg_ref_dict:
                    MappingRecord_dict[read_id_base].r1_ctg_ref_dict[ref_id] = {ref_pos: cigar}
                else:
                    MappingRecord_dict[read_id_base].r1_ctg_ref_dict[ref_id][ref_pos] = cigar

            if read_strand == '2':
                if ref_id not in MappingRecord_dict[read_id_base].r2_ctg_ref_dict:
                    MappingRecord_dict[read_id_base].r2_ctg_ref_dict[ref_id] = {ref_pos: cigar}
                else:
                    MappingRecord_dict[read_id_base].r2_ctg_ref_dict[ref_id][ref_pos] = cigar


##################################################### parse MappingRecord_dict ####################################################

processed_num = 0
for each_mp in MappingRecord_dict:

    # print('%s\tr1_ctg_ref_dict\t%s'       % (each_mp, current_mp_record.r1_ctg_ref_dict))
    # print('%s\tr2_ctg_ref_dict\t%s'       % (each_mp, current_mp_record.r2_ctg_ref_dict))

    current_mp_record = MappingRecord_dict[each_mp]
    refs_to_ignore_ctg = set()

    ########## filter r1 ctg refs ##########

    r1_ctg_refs_passed_qc = {}
    for r1_ctg_ref in current_mp_record.r1_ctg_ref_dict:
        r1_ctg_ref_matched_pos_dict = current_mp_record.r1_ctg_ref_dict[r1_ctg_ref]

        # one read need to mapped to one ctg only for one time
        if len(r1_ctg_ref_matched_pos_dict) > 1:
            refs_to_ignore_ctg.add(r1_ctg_ref)
        else:
            r1_ctg_ref_pos = list(r1_ctg_ref_matched_pos_dict.keys())[0]
            r1_ctg_ref_cigar = r1_ctg_ref_matched_pos_dict[r1_ctg_ref_pos]
            r1_ctg_ref_cigar_splitted = cigar_splitter(r1_ctg_ref_cigar)

            # check both end clip
            both_end_clp = check_both_ends_clipping(r1_ctg_ref_cigar_splitted)
            if both_end_clp is True:
                refs_to_ignore_ctg.add(r1_ctg_ref)
            else:
                # check mismatch
                r1_aligned_len_ctg, r1_aligned_pct_ctg, r1_clipping_len_ctg, r1_clipping_pct_ctg, r1_mismatch_pct_ctg = get_cigar_stats(r1_ctg_ref_cigar_splitted)
                if r1_mismatch_pct_ctg > mismatch_cutoff:
                    refs_to_ignore_ctg.add(r1_ctg_ref)
                else:
                    # check aligned length
                    if r1_aligned_len_ctg < min_M_len_ctg:
                        refs_to_ignore_ctg.add(r1_ctg_ref)
                    else:
                        # check if clp in the middle
                        clip_in_middle = False
                        if ('S' in r1_ctg_ref_cigar) or ('s' in r1_ctg_ref_cigar):
                            clip_in_middle = True
                            if (r1_ctg_ref_cigar_splitted[0][-1] in ['S', 's']) and (r1_ctg_ref_pos == 1):
                                clip_in_middle = False
                            if (r1_ctg_ref_cigar_splitted[-1][-1] in ['S', 's']):
                                if (r1_ctg_ref_pos + r1_aligned_len_ctg - 1) == ctg_len_dict[r1_ctg_ref]:
                                    clip_in_middle = False

                        # exclude the ref if clp in the middle is True
                        if clip_in_middle is True:
                            refs_to_ignore_ctg.add(r1_ctg_ref)
                        else:
                            r1_ctg_refs_passed_qc[r1_ctg_ref] = [r1_ctg_ref_cigar]

    ########## filter r2 ctg refs ##########

    r2_ctg_refs_passed_qc = {}
    for r2_ctg_ref in current_mp_record.r2_ctg_ref_dict:
        r2_ctg_ref_matched_pos_dict = current_mp_record.r2_ctg_ref_dict[r2_ctg_ref]

        # one read need to mapped to one ctg only for one time
        if len(r2_ctg_ref_matched_pos_dict) > 1:
            refs_to_ignore_ctg.add(r2_ctg_ref)
        else:
            r2_ctg_ref_pos = list(r2_ctg_ref_matched_pos_dict.keys())[0]
            r2_ctg_ref_cigar = r2_ctg_ref_matched_pos_dict[r2_ctg_ref_pos]
            r2_ctg_ref_cigar_splitted = cigar_splitter(r2_ctg_ref_cigar)

            # check both end clip
            both_end_clp = check_both_ends_clipping(r2_ctg_ref_cigar_splitted)
            if both_end_clp is True:
                refs_to_ignore_ctg.add(r2_ctg_ref)
            else:
                # check mismatch
                r2_aligned_len_ctg, r2_aligned_pct_ctg, r2_clipping_len_ctg, r2_clipping_pct_ctg, r2_mismatch_pct_ctg = get_cigar_stats(r2_ctg_ref_cigar_splitted)
                if r2_mismatch_pct_ctg > mismatch_cutoff:
                    refs_to_ignore_ctg.add(r2_ctg_ref)
                else:
                    # check aligned length
                    if r2_aligned_len_ctg < min_M_len_ctg:
                        refs_to_ignore_ctg.add(r2_ctg_ref)
                    else:
                        # check if clp in the middle
                        clip_in_middle = False
                        if ('S' in r2_ctg_ref_cigar) or ('s' in r2_ctg_ref_cigar):
                            clip_in_middle = True
                            if (r2_ctg_ref_cigar_splitted[0][-1] in ['S', 's']) and (r2_ctg_ref_pos == 1):
                                clip_in_middle = False
                            if (r2_ctg_ref_cigar_splitted[-1][-1] in ['S', 's']):
                                if (r2_ctg_ref_pos + r2_aligned_len_ctg - 1) == ctg_len_dict[r2_ctg_ref]:
                                    clip_in_middle = False

                        # exclude the ref if clp in the middle is True
                        if clip_in_middle is True:
                            refs_to_ignore_ctg.add(r2_ctg_ref)
                        else:
                            r2_ctg_refs_passed_qc[r2_ctg_ref] = [r2_ctg_ref_cigar]

    ####################################################################################################

    r1_ctg_refs_no_ignored = {key: value for key, value in r1_ctg_refs_passed_qc.items() if key not in refs_to_ignore_ctg}
    r2_ctg_refs_no_ignored = {key: value for key, value in r2_ctg_refs_passed_qc.items() if key not in refs_to_ignore_ctg}

    # no mate matched to ctg
    if (len(r1_ctg_refs_no_ignored) == 0) and (len(r2_ctg_refs_no_ignored) == 0):
        pass

    # only r1 matched to ctg
    elif (len(r1_ctg_refs_no_ignored) > 0) and (len(r2_ctg_refs_no_ignored) == 0):
        MappingRecord_dict[each_mp].matched_to_ctg = True
        MappingRecord_dict[each_mp].r1_ctg_refs_no_ignored = r1_ctg_refs_no_ignored

    # only r2 matched to ctg
    elif (len(r1_ctg_refs_no_ignored) == 0) and (len(r2_ctg_refs_no_ignored) > 0):
        MappingRecord_dict[each_mp].matched_to_ctg = True
        MappingRecord_dict[each_mp].r2_ctg_refs_no_ignored = r2_ctg_refs_no_ignored

    # both r1 and r2 matched to ctg
    else:
        shared_ctg_ref_set = {key: [r1_ctg_refs_no_ignored[key][0], r2_ctg_refs_no_ignored[key][0]] for key in set(r1_ctg_refs_no_ignored).intersection(set(r2_ctg_refs_no_ignored))}
        if len(shared_ctg_ref_set) > 0:
            MappingRecord_dict[each_mp].matched_to_ctg = True
            MappingRecord_dict[each_mp].shared_ctg_refs_no_ignored = shared_ctg_ref_set

    processed_num += 1
    if (processed_num % report_interval == 0):
        processed_pct = processed_num*100/len(MappingRecord_dict)
        processed_pct = float("{0:.2f}".format(processed_pct))
        print('Processed %sk (%s%s)' % (int(processed_num/1000), processed_pct, '%'))


##################################################### get linkages from MappingRecord_dict #####################################################

marker_to_ctg_linkage_cigar_dict_16s_side = {}
marker_to_ctg_linkage_cigar_dict_ctg_side = {}
marker_to_ctg_linkage_num_dict = {}
marker_to_ctg_linking_reads_dict = {}
for qualified_read in MappingRecord_dict:

    read_mr         = MappingRecord_dict[qualified_read]
    r1_16s_refs     = read_mr.r1_16s_refs_no_ignored
    r2_16s_refs     = read_mr.r2_16s_refs_no_ignored
    shared_16s_refs = read_mr.shared_16s_refs_no_ignored
    r1_ctg_refs     = read_mr.r1_ctg_refs_no_ignored
    r2_ctg_refs     = read_mr.r2_ctg_refs_no_ignored
    shared_ctg_refs = read_mr.shared_ctg_refs_no_ignored

    # combine dict
    combined_16s_refs_dict = {**r1_16s_refs, **r2_16s_refs, **shared_16s_refs}
    combined_ctg_refs_dict = {**r1_ctg_refs, **r2_ctg_refs, **shared_ctg_refs}

    if (len(combined_16s_refs_dict) > 0) and (len(combined_ctg_refs_dict) > 0):
        for each_16s_ref in combined_16s_refs_dict:
            current_cigar_list_16s_side = combined_16s_refs_dict[each_16s_ref]
            for each_ctg_ref in combined_ctg_refs_dict:
                current_cigar_list_ctg_side = combined_ctg_refs_dict[each_ctg_ref]
                marker_to_ctg_key = '%s%s%s' % (each_16s_ref, marker_to_ctg_gnm_Key_connector, each_ctg_ref)

                if marker_to_ctg_key not in marker_to_ctg_linkage_num_dict:
                    marker_to_ctg_linkage_num_dict[marker_to_ctg_key] = 1
                    marker_to_ctg_linking_reads_dict[marker_to_ctg_key] = {qualified_read}
                    marker_to_ctg_linkage_cigar_dict_16s_side[marker_to_ctg_key] = current_cigar_list_16s_side
                    marker_to_ctg_linkage_cigar_dict_ctg_side[marker_to_ctg_key] = current_cigar_list_ctg_side
                else:
                    marker_to_ctg_linkage_num_dict[marker_to_ctg_key] += 1
                    marker_to_ctg_linking_reads_dict[marker_to_ctg_key].add(qualified_read)
                    marker_to_ctg_linkage_cigar_dict_16s_side[marker_to_ctg_key] += current_cigar_list_16s_side
                    marker_to_ctg_linkage_cigar_dict_ctg_side[marker_to_ctg_key] += current_cigar_list_ctg_side


# calculate clp_pct_ratio
rd1_clp_pct_diff_txt_handle = open(rd1_clp_pct_diff_txt, 'w')
rd1_clp_pct_diff_txt_to_ignore_handle = open(rd1_clp_pct_diff_txt_to_ignore, 'w')
rd1_clp_pct_diff_txt_handle.write('Marker\tGenome\tContig\tclp_pct_ctg\tclp_pct_16s\tRatio\n')
rd1_clp_pct_diff_txt_to_ignore_handle.write('Marker\tGenome\tContig\tclp_pct_ctg\tclp_pct_16s\tRatio\n')
linkages_to_ignore = set()
for each_link in marker_to_ctg_linkage_num_dict:
    if marker_to_ctg_linkage_num_dict[each_link] >= 3:
        each_link_split = each_link.split(marker_to_ctg_gnm_Key_connector)
        marker_id = each_link_split[0]
        gnm_id = each_link_split[1].split(gnm_to_ctg_connector)[0]
        ctg_id = each_link_split[1].split(gnm_to_ctg_connector)[1]

        linkage_cigar_16s_side_all = marker_to_ctg_linkage_cigar_dict_16s_side[each_link]
        linkage_cigar_ctg_side_all = marker_to_ctg_linkage_cigar_dict_ctg_side[each_link]
        linkage_cigar_16s_side_clp = [i for i in linkage_cigar_16s_side_all if (('S' in i) or ('s' in i))]
        linkage_cigar_ctg_side_clp = [i for i in linkage_cigar_ctg_side_all if (('S' in i) or ('s' in i))]

        clp_pct_16s_side = 'na'
        if len(linkage_cigar_16s_side_all) > 0:
            clp_pct_16s_side = len(linkage_cigar_16s_side_clp) * 100 / len(linkage_cigar_16s_side_all)
            clp_pct_16s_side = float("{0:.2f}".format(clp_pct_16s_side))

        clp_pct_ctg_side = 'na'
        if len(linkage_cigar_ctg_side_all) > 0:
            clp_pct_ctg_side = len(linkage_cigar_ctg_side_clp) * 100 / len(linkage_cigar_ctg_side_all)
            clp_pct_ctg_side = float("{0:.2f}".format(clp_pct_ctg_side))

        clp_pct_ratio = 'na'
        if (clp_pct_16s_side != 'na') and (clp_pct_ctg_side != 'na'):
            if clp_pct_16s_side > 0:
                clp_pct_ratio = float("{0:.2f}".format(clp_pct_ctg_side / clp_pct_16s_side))

        to_ignore = False
        if clp_pct_ratio != 'na':
            if (clp_pct_ctg_side >= 65) and (clp_pct_ratio >= 5):
                to_ignore = True

        if to_ignore is True:
            linkages_to_ignore.add(each_link)
            rd1_clp_pct_diff_txt_to_ignore_handle.write('%s\t%s\t%s\t%s(%s/%s)\t%s(%s/%s)\t%s\n' % (marker_id, gnm_id, ctg_id, clp_pct_ctg_side, len(linkage_cigar_ctg_side_clp), len(linkage_cigar_ctg_side_all), clp_pct_16s_side, len(linkage_cigar_16s_side_clp), len(linkage_cigar_16s_side_all), clp_pct_ratio))
        else:
            rd1_clp_pct_diff_txt_handle.write('%s\t%s\t%s\t%s(%s/%s)\t%s(%s/%s)\t%s\n' % (marker_id, gnm_id, ctg_id, clp_pct_ctg_side, len(linkage_cigar_ctg_side_clp), len(linkage_cigar_ctg_side_all), clp_pct_16s_side, len(linkage_cigar_16s_side_clp), len(linkage_cigar_16s_side_all), clp_pct_ratio))
rd1_clp_pct_diff_txt_handle.close()
rd1_clp_pct_diff_txt_to_ignore_handle.close()

# ignore linkages if did not pass the clp pct ratio test
marker_to_ctg_linkage_num_dict_min3_passed_ratio_check = {}
for each_link in marker_to_ctg_linkage_num_dict:
    if marker_to_ctg_linkage_num_dict[each_link] >= 3:
        if each_link not in linkages_to_ignore:
            marker_to_ctg_linkage_num_dict_min3_passed_ratio_check[each_link] = marker_to_ctg_linkage_num_dict[each_link]

# get number of linkages at genome level
marker_to_gnm_link_num = {}
for each_marker_to_ctg_key in marker_to_ctg_linkage_num_dict_min3_passed_ratio_check:
    marker_id = each_marker_to_ctg_key.split(marker_to_ctg_gnm_Key_connector)[0]
    ctg_id = each_marker_to_ctg_key.split(marker_to_ctg_gnm_Key_connector)[1]
    gnm_id = ctg_id.split(gnm_to_ctg_connector)[0]
    marker_to_gnm_key = '%s%s%s' % (marker_id, marker_to_ctg_gnm_Key_connector, gnm_id)
    if marker_to_gnm_key not in marker_to_gnm_link_num:
        marker_to_gnm_link_num[marker_to_gnm_key] = marker_to_ctg_linkage_num_dict_min3_passed_ratio_check[each_marker_to_ctg_key]
    else:
        marker_to_gnm_link_num[marker_to_gnm_key] += marker_to_ctg_linkage_num_dict_min3_passed_ratio_check[each_marker_to_ctg_key]

# write out linkages at genome level
sankey_file_in_handle = open(link_stats_combined, 'w')
sankey_file_in_handle.write('MarkerGene,GenomicSeq,Number\n')
for each_linkage in marker_to_gnm_link_num:
    sankey_file_in_handle.write('MarkerGene__%s,GenomicSeq__%s,%s\n' % (
    each_linkage.split(marker_to_ctg_gnm_Key_connector)[0], each_linkage.split(marker_to_ctg_gnm_Key_connector)[1],
    marker_to_gnm_link_num[each_linkage]))
sankey_file_in_handle.close()

# write out linking reads
linking_reads_rd1_handle = open(linking_reads_rd1, 'w')
for each_link in marker_to_ctg_linkage_num_dict:
    marker_id = each_link.split(marker_to_ctg_gnm_Key_connector)[0]
    ctg_id = each_link.split(marker_to_ctg_gnm_Key_connector)[1]
    gnm_id = ctg_id.split(gnm_to_ctg_connector)[0]
    current_linking_reads = marker_to_ctg_linking_reads_dict[each_link]
    linking_reads_rd1_handle.write('%s\t%s\n' % (each_link, ','.join(current_linking_reads)))
linking_reads_rd1_handle.close()






def check_both_ends_clipping(cigar_splitted):

    both_ends_clipping = False
    if len(cigar_splitted) >= 3:
        if (cigar_splitted[0][-1] in ['S', 's']) and (cigar_splitted[-1][-1] in ['S', 's']):
            both_ends_clipping = True

    return both_ends_clipping


def cigar_splitter(cigar):

    # get the position of letters
    letter_pos_list = []
    n = 0
    for each_element in cigar:
        if (each_element.isalpha() is True) or (each_element == '='):
            letter_pos_list.append(n)
        n += 1

    # split cigar
    index = 0
    cigar_splitted = []
    while index <= len(letter_pos_list) - 1:
        if index == 0:
            cigar_splitted.append(cigar[:(letter_pos_list[index] + 1)])
        else:
            cigar_splitted.append(cigar[(letter_pos_list[index - 1] + 1):(letter_pos_list[index] + 1)])
        index += 1

    return cigar_splitted


def get_cigar_stats(cigar_splitted):

    # aligned_len: M I X =
    # clipping_len: S
    # mismatch_len: X I D
    # mismatch_pct = mismatch_len / aligned_len
    # aligned_pct  = aligned_len  / (aligned_len + clipping_len)
    # clipping_pct = clipping_len / (aligned_len + clipping_len)

    aligned_len = 0
    clipping_len = 0
    mismatch_len = 0
    for each_part in cigar_splitted:
        each_part_len = int(each_part[:-1])
        each_part_cate = each_part[-1]

        # get aligned_len
        if each_part_cate in {'M', 'm', 'I', 'i', 'X', 'x', '='}:
            aligned_len += each_part_len

        # get clipping_len
        if each_part_cate in ['S', 's']:
            clipping_len += each_part_len

        # get mismatch_len
        if each_part_cate in {'I', 'i', 'X', 'x', 'D', 'd'}:
            mismatch_len += each_part_len

    aligned_pct  = float("{0:.2f}".format(aligned_len * 100 / (aligned_len + clipping_len)))
    clipping_pct = float("{0:.2f}".format(clipping_len * 100 / (aligned_len + clipping_len)))
    mismatch_pct = float("{0:.2f}".format(mismatch_len * 100 / (aligned_len)))

    return aligned_len, aligned_pct, clipping_len, clipping_pct, mismatch_pct


def keep_best_matches_in_sam_keep_short_M(sam_in, min_M_len, sam_out):

    # get read_to_cigar_dict
    read_to_cigar_dict = {}
    for each_line in open(sam_in):
        each_line_split = each_line.strip().split('\t')
        if not each_line.startswith('@'):
            read_id = each_line_split[0]
            cigar = each_line_split[5]
            if cigar != '*':
                both_ends_clp = check_both_ends_clipping(cigar_splitter(cigar))
                if both_ends_clp is False:
                    if read_id not in read_to_cigar_dict:
                        read_to_cigar_dict[read_id] = {cigar}
                    else:
                        read_to_cigar_dict[read_id].add(cigar)

    # get min_mismatch for each read
    read_min_mismatch_dict = {}
    for each_read in read_to_cigar_dict:
        read_mismatch_set_all_M = set()
        read_mismatch_set_long_M = set()
        for each_cigar in read_to_cigar_dict[each_read]:
            aligned_len, aligned_pct, clipping_len, clipping_pct, mismatch_pct = get_cigar_stats(cigar_splitter(each_cigar))
            read_mismatch_set_all_M.add(mismatch_pct)
            if aligned_len >= min_M_len:
                read_mismatch_set_long_M.add(mismatch_pct)
        read_min_mismatch = min(read_mismatch_set_all_M)
        if len(read_mismatch_set_long_M) > 0:
            read_min_mismatch = min(read_mismatch_set_long_M)
        read_min_mismatch_dict[each_read] = read_min_mismatch

    sam_file_best_match_handle = open(sam_out, 'w')
    for each_line in open(sam_in):
        if each_line.startswith('@'):
            sam_file_best_match_handle.write(each_line)
        else:
            each_line_split = each_line.strip().split('\t')
            cigar = each_line_split[5]
            if cigar == '*':
                sam_file_best_match_handle.write(each_line)
            else:
                read_id = each_line_split[0]
                cigar_split = cigar_splitter(cigar)
                both_ends_clp = check_both_ends_clipping(cigar_splitter(cigar))
                if both_ends_clp is False:
                    aligned_len, aligned_pct, clipping_len, clipping_pct, mismatch_pct = get_cigar_stats(cigar_split)
                    if mismatch_pct <= (read_min_mismatch_dict[read_id] * 1.5):
                        sam_file_best_match_handle.write(each_line)
    sam_file_best_match_handle.close()


sam_in      = '/Users/songweizhi/Desktop/tunning/sub_all.sam'
sam_out     = '/Users/songweizhi/Desktop/tunning/sub_all_best.sam'
min_M_len   = 75

keep_best_matches_in_sam_keep_short_M(sam_in, min_M_len, sam_out)



'''
cami_hc_S5_42397362.1	{'cami_hc_SILVA138_id99_50_subsample_50_1792'}
cami_hc_S5_42397362.1	345	cami_hc_SILVA138_id99_75_subsample_75_2741	1	11	24S13=1X62=	=	1	0	CAGTTTTTCAACAGATTTTGTTGGAGAGTTTGATCCTCGCTCAGGACGAACGCTGGCGGCGTGCTTAACACATGCAAGTCGAGCGGAAAGGCCCTTCGGG	IIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIII	AS:i:144	XN:i:0	XM:i:1	XO:i:0	XG:i:0	NM:i:1	MD:Z:13G62	YT:Z:UP
cami_hc_S5_42397362.1	345	cami_hc_SILVA138_id99_50_subsample_10_876	1	11	24S13=1X62=	=	1	0	CAGTTTTTCAACAGATTTTGTTGGAGAGTTTGATCCTCGCTCAGGACGAACGCTGGCGGCGTGCTTAACACATGCAAGTCGAGCGGAAAGGCCCTTCGGG	IIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIII	AS:i:144	XN:i:0	XM:i:1	XO:i:0	XG:i:0	NM:i:1	MD:Z:13G62	YT:Z:UP
cami_hc_S5_42397362.1	89	cami_hc_SILVA138_id99_50_subsample_50_1792	1	11	16S21=1X62=	=	1	0	CAGTTTTTCAACAGATTTTGTTGGAGAGTTTGATCCTCGCTCAGGACGAACGCTGGCGGCGTGCTTAACACATGCAAGTCGAGCGGAAAGGCCCTTCGGG	IIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIII	AS:i:160	XN:i:0	XM:i:1	XO:i:0	XG:i:0	NM:i:1	MD:Z:21G62	YT:Z:UP
'''


import seaborn as sns
import plotly.graph_objects as go


def get_sankey_plot(node_list, source_list, target_list, value_list, color_list, plot_title, plot_height, output_html):

    node_index_dict = {y: x for x, y in enumerate(node_list)}
    source_index = [node_index_dict[x] for x in source_list]
    target_index = [node_index_dict[x] for x in target_list]

    # https://anvil.works/docs/api/plotly.graph_objs.sankey
    fig = go.Figure(data=[go.Sankey(node=dict(label=node_list,  # line=0,
                                              pad=5,  # space between node
                                              thickness=12,  # node width
                                              line=dict(width=0)),  # set width of node border to 0
                                    link=dict(source=source_index,
                                              target=target_index,
                                              value=value_list,
                                              color=color_list))])

    fig.update_layout(autosize=False, width=1200, height=plot_height, margin=dict(l=50, r=50, b=50, t=125), paper_bgcolor="white", title=plot_title)
    fig.update_traces(textfont_size=11)
    fig.write_html(output_html)


def sankey_linkages(combined_linkage_file_ctg_level, linkage_plot_rd1_html, linkage_plot_rd2_html):

    dict_for_sankey_key_connector   = '___X___'

    linkage_num_dict_rd1 = {}
    linkage_num_dict_rd2 = {}
    ctg_to_gnm_dict_rd1 = {}
    ctg_to_gnm_dict_rd2 = {}
    node_set_rd1 = set()
    node_set_rd2 = set()
    genome_set_rd1 = set()
    genome_set_rd2 = set()
    contig_set_rd1 = set()
    contig_set_rd2 = set()
    marker_gene_set_rd1 = set()
    marker_gene_set_rd2 = set()
    for each_linkage in open(combined_linkage_file_ctg_level):
        if not each_linkage.startswith('Marker___Genome(total)	Contig	Paired	Clipping	Overlapped	Step'):
            each_linkage_split = each_linkage.strip().split('\t')

            marker_id = each_linkage_split[0].split('___')[0]
            gnm_id = each_linkage_split[0].split('___')[1].split('(')[0]
            ctg_id = each_linkage_split[1]
            total_link_num = int(each_linkage_split[2]) + int(each_linkage_split[3]) + int(each_linkage_split[4])
            marker_to_ctg_key = '%s%s%s' % (marker_id, dict_for_sankey_key_connector, ctg_id)
            ctg_to_gnm_key = '%s%s%s' % (ctg_id, dict_for_sankey_key_connector, gnm_id)

            if each_linkage_split[5] == 'S1':
                genome_set_rd1.add(gnm_id)
                contig_set_rd1.add(ctg_id)
                marker_gene_set_rd1.add(marker_id)
                node_set_rd1.add(marker_id)
                node_set_rd1.add(ctg_id)
                node_set_rd1.add(gnm_id)

                if ctg_id not in ctg_to_gnm_dict_rd1:
                    ctg_to_gnm_dict_rd1[ctg_id] = gnm_id

                if marker_to_ctg_key not in linkage_num_dict_rd1:
                    linkage_num_dict_rd1[marker_to_ctg_key] = total_link_num
                else:
                    linkage_num_dict_rd1[marker_to_ctg_key] += total_link_num

                if ctg_to_gnm_key not in linkage_num_dict_rd1:
                    linkage_num_dict_rd1[ctg_to_gnm_key] = total_link_num
                else:
                    linkage_num_dict_rd1[ctg_to_gnm_key] += total_link_num

            if each_linkage_split[5] == 'S2':
                genome_set_rd2.add(gnm_id)
                contig_set_rd2.add(ctg_id)
                marker_gene_set_rd2.add(marker_id)
                node_set_rd2.add(marker_id)
                node_set_rd2.add(ctg_id)
                node_set_rd2.add(gnm_id)
                if ctg_id not in ctg_to_gnm_dict_rd2:
                    ctg_to_gnm_dict_rd2[ctg_id] = gnm_id

                if marker_to_ctg_key not in linkage_num_dict_rd2:
                    linkage_num_dict_rd2[marker_to_ctg_key] = total_link_num
                else:
                    linkage_num_dict_rd2[marker_to_ctg_key] += total_link_num

                if ctg_to_gnm_key not in linkage_num_dict_rd2:
                    linkage_num_dict_rd2[ctg_to_gnm_key] = total_link_num
                else:
                    linkage_num_dict_rd2[ctg_to_gnm_key] += total_link_num

    source_list_rd1 = []
    target_list_rd1 = []
    value_list_rd1 = []
    for each_rd1_linkage in linkage_num_dict_rd1:
        each_rd1_linkage_split = each_rd1_linkage.split(dict_for_sankey_key_connector)
        source_list_rd1.append(each_rd1_linkage_split[0])
        target_list_rd1.append(each_rd1_linkage_split[1])
        value_list_rd1.append(linkage_num_dict_rd1[each_rd1_linkage])

    source_list_rd2 = []
    target_list_rd2 = []
    value_list_rd2 = []
    for each_rd2_linkage in linkage_num_dict_rd2:
        each_rd2_linkage_split = each_rd2_linkage.split(dict_for_sankey_key_connector)
        source_list_rd2.append(each_rd2_linkage_split[0])
        target_list_rd2.append(each_rd2_linkage_split[1])
        value_list_rd2.append(linkage_num_dict_rd2[each_rd2_linkage])

    gnm_color_list_rd1 = sns.color_palette('tab20', len(genome_set_rd1)).as_hex()
    gnm_color_list_rd2 = sns.color_palette('tab20', len(genome_set_rd2)).as_hex()

    genome_to_color_dict_rd1 = {gnm: color for gnm, color in zip(genome_set_rd1, gnm_color_list_rd1)}
    genome_to_color_dict_rd2 = {gnm: color for gnm, color in zip(genome_set_rd2, gnm_color_list_rd2)}

    color_list_rd1 = []
    for each_target in target_list_rd1:
        if each_target in genome_to_color_dict_rd1:
            color_list_rd1.append(genome_to_color_dict_rd1[each_target])
        else:
            target_genome = ctg_to_gnm_dict_rd1[each_target]
            color_list_rd1.append(genome_to_color_dict_rd1[target_genome])

    color_list_rd2 = []
    for each_target in target_list_rd2:
        if each_target in genome_to_color_dict_rd2:
            color_list_rd2.append(genome_to_color_dict_rd2[each_target])
        else:
            target_genome = ctg_to_gnm_dict_rd2[each_target]
            color_list_rd2.append(genome_to_color_dict_rd2[target_genome])

    node_list_rd1 = sorted([i for i in node_set_rd1])
    node_list_rd2 = sorted([i for i in node_set_rd2])

    plot_title_text_rd1 = 'MarkerMAG detected linkages (round 1)<br>Number of linked genomes: %s<br>Number of linked markers: %s' % (len(genome_set_rd1), len(marker_gene_set_rd1))
    plot_title_text_rd2 = 'MarkerMAG detected linkages (round 2)<br>Number of linked genomes: %s<br>Number of linked markers: %s' % (len(genome_set_rd2), len(marker_gene_set_rd2))

    plot_height_rd1 = 500 if max([len(contig_set_rd1), len(marker_gene_set_rd1)]) <= 25 else max([len(contig_set_rd1), len(marker_gene_set_rd1)]) * 32
    plot_height_rd2 = 500 if max([len(contig_set_rd2), len(marker_gene_set_rd2)]) <= 25 else max([len(contig_set_rd2), len(marker_gene_set_rd2)]) * 32

    plot_title_dict_rd1 = dict(text=plot_title_text_rd1, x=0.05, y=(1-(50/plot_height_rd1)))
    plot_title_dict_rd2 = dict(text=plot_title_text_rd2, x=0.05, y=(1-(50/plot_height_rd2)))

    get_sankey_plot(node_list_rd1, source_list_rd1, target_list_rd1, value_list_rd1, color_list_rd1, plot_title_dict_rd1, plot_height_rd1, linkage_plot_rd1_html)
    get_sankey_plot(node_list_rd2, source_list_rd2, target_list_rd2, value_list_rd2, color_list_rd2, plot_title_dict_rd2, plot_height_rd2, linkage_plot_rd2_html)


combined_linkage_file_ctg_level = '/Users/songweizhi/Desktop/GI_0406_super_specific_identified_linkages_contig_level.txt'
linkage_plot_rd1_html           = '/Users/songweizhi/Desktop/GI_rd1.html'
linkage_plot_rd2_html           = '/Users/songweizhi/Desktop/GI_rd2.html'

sankey_linkages(combined_linkage_file_ctg_level, linkage_plot_rd1_html, linkage_plot_rd2_html)
